#!/usr/bin/env python
"""
M2Crypto SSLTransport Library
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import socket
from M2Crypto import SSL, threading as M2Threading
from M2Crypto.SSL.Checker import SSLVerificationError

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.DISET.private.Transports.BaseTransport import BaseTransport
from DIRAC.Core.DISET.private.Transports.SSL.M2Utils import getM2SSLContext, getM2PeerInfo

from DIRAC.Core.DISET import DEFAULT_CONNECTION_TIMEOUT, DEFAULT_RPC_TIMEOUT

# TODO: For now we have to set an environment variable for proxy support in OpenSSL
# Eventually we may need to add API support for this to M2Crypto...
os.environ["OPENSSL_ALLOW_PROXY_CERTS"] = "1"
M2Threading.init()

# TODO: CRL checking should be implemented but this will require support adding
# to M2Crypto: Quite a few functions will need mapping through from OpenSSL to
# allow the CRL stack to be set on the X509 CTX used for verification.

# TODO: Log useful messages to the logger


class SSLTransport(BaseTransport):
    """SSL Transport implementation using the M2Crypto library."""

    # This name is the same as BaseClient,
    # and is used a bit everywhere, so it should be factorized out
    # eventually
    KW_TIMEOUT = "timeout"

    def __getConnection(self):
        """Helper function to get a connection object,
        Tries IPv6 (AF_INET6) first, then falls back to IPv4 (AF_INET).
        """
        try:
            conn = SSL.Connection(self.__ctx, family=socket.AF_INET6)
        except socket.error:
            # Maybe no IPv6 support? Try IPv4 only socket.
            conn = SSL.Connection(self.__ctx, family=socket.AF_INET)
        return conn

    def __init__(self, *args, **kwargs):
        """Create an SSLTransport object, parameters are the same
        as for other transports. If ctx is specified (as an instance of
        SSL.Context) then use that rather than creating a new context.

        kwargs can contain all the parameters defined in BaseClient,
        in particular timeout
        """
        # The thread init of M2Crypto is not really thread safe.
        # So we put it a second time
        M2Threading.init()
        self.remoteAddress = None
        self.peerCredentials = {}

        # The timeout used here is different from what it was in pyGSI.
        # It is to be understood here as the timeout for socket operations
        # involved in the RPC call, but NOT the establishment of the connection,
        # for which there is a different timeout.
        #
        # The timeout management of pyGSI was a bit off.
        # This is proven by that type of trace (look at the timestamp):
        #
        # 2020-07-16 09:48:55 UTC dirac-proxy-init [140013698656064] DEBUG: Connection timeout set to:  1
        # 2020-07-16 09:58:55 UTC dirac-proxy-init [140013698656064] WARN: Issue getting socket:
        #

        self.__timeout = kwargs.get(SSLTransport.KW_TIMEOUT, DEFAULT_RPC_TIMEOUT)

        self.__locked = False  # We don't support locking, so this is always false.

        # If not specified in the arguments (never is in DIRAC code...)
        # and we are setting up a server listing connection, set the accepted
        # ssl methods and ciphers
        if kwargs.get("bServerMode"):
            if "sslMethods" not in kwargs:
                kwargs["sslMethods"] = os.environ.get("DIRAC_M2CRYPTO_SSL_METHODS")
            if "sslCiphers" not in kwargs:
                kwargs["sslCiphers"] = os.environ.get("DIRAC_M2CRYPTO_SSL_CIPHERS")

        self.__ctx = kwargs.pop("ctx", None)
        if not self.__ctx:
            self.__ctx = getM2SSLContext(**kwargs)

        # Note that kwargs is already kept in BaseTransport
        # as self.extraArgsDict, but at least I am sure that
        # self.__kwargs will never be modified
        self.__kwargs = kwargs

        BaseTransport.__init__(self, *args, **kwargs)

    def setSocketTimeout(self, timeout):
        """Set the timeout for RPC calls.

        .. warning: This needs to be called before initAsClient.
          It is used as a timeout for RPC calls, not connection.

        :param timeout: timeout for socket operation in seconds

        """
        self.__timeout = timeout

    def initAsClient(self):
        """Prepare this client socket for use."""
        if self.serverMode():
            raise RuntimeError("SSLTransport is in server mode.")

        error = None
        host, port = self.stServerAddress

        # The following piece of code was inspired by the python socket documentation
        # as well as the implementation of M2Crypto.httpslib.HTTPSConnection

        # We ignore the returned sockaddr because SSL.Connection.connect needs
        # a host name.
        addrInfoList = socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
        for (family, _socketType, _proto, _canonname, _socketAddress) in addrInfoList:

            try:
                self.oSocket = SSL.Connection(self.__ctx, family=family)

                # First set a short connection timeout, that will trigger
                # during blocking operations (read/write) use to
                # establish the SSL connection
                self.oSocket.settimeout(DEFAULT_CONNECTION_TIMEOUT)

                # Enable keepAlive, with default options
                # (see more comments about keepalive in :py:meth:`.acceptConnection`)
                self.oSocket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, True)

                # set SNI server name since we know it at this point
                self.oSocket.set_tlsext_host_name(host)

                self.oSocket.connect((host, port))

                # Once the connection is established, we can use the timeout
                # asked for RPC
                self.oSocket.settimeout(self.__timeout)

                self.remoteAddress = self.oSocket.getpeername()

                return S_OK()
            # warning: do NOT catch SSL related error here
            # They should be propagated upwards and caught by the BaseClient
            # not to enter the retry loop
            except socket.error as e:
                error = "%s:%s" % (e, repr(e))

                if self.oSocket is not None:
                    self.close()

        return S_ERROR(error)

    def initAsServer(self):
        """Prepare this server socket for use."""
        if not self.serverMode():
            raise RuntimeError("SSLTransport is in client mode.")
        # Before getting the connection object, we need to set
        # a server session ID in the context
        host = self.stServerAddress[0]
        port = self.stServerAddress[1]
        self.__ctx.set_session_id_ctx(("DIRAC-%s-%s" % (host, port)).encode())
        self.oSocket = self.__getConnection()
        # Make sure reuse address is set correctly
        if self.bAllowReuseAddress:
            param = 1
        else:
            param = 0

        self.oSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, param)

        self.oSocket.bind(self.stServerAddress)
        self.oSocket.listen(self.iListenQueueSize)
        return S_OK()

    def close(self):
        # pylint: disable=line-too-long
        """Close this socket."""

        if self.oSocket:
            # TL;DR:
            # Do NOT touch that method
            #
            # Surprisingly (to me at least), M2Crypto does not close
            # the underlying socket when calling SSL.Connection.close
            # It only does it when the garbage collector kicks in (see ~M2Crypto.SSL.Connection.Connection.__del__)
            # If the socket is not closed, the connection may hang forever.
            #
            # Thus, we are setting self.oSocket to None to allow the GC to do the work, but since we are not sure
            # that it will run, we anyway force the connection to be closed
            #
            # However, we should close the underlying socket only after SSL was shutdown properly.
            # This is because OpenSSL `ssl3_shutdown` (see callstack below) may still read some data
            # (see https://github.com/openssl/openssl/blob/master/ssl/s3_lib.c#L4509)::
            #
            #
            # 1  0x00007fffe9d48fc0 in sock_read () from /lib/libcrypto.so.1.0.0
            # 2  0x00007fffe9d46e83 in BIO_read () from /lib/libcrypto.so.1.0.0
            # 3  0x00007fffe9eab9dd in ssl3_read_n () from /lib/libssl.so.1.0.0
            # 4  0x00007fffe9ead216 in ssl3_read_bytes () from /lib/libssl.so.1.0.0
            # 5  0x00007fffe9ea999c in ssl3_shutdown () from /lib/libssl.so.1.0.0
            # 6  0x00007fffe9ed4f93 in ssl_free () from /lib/libssl.so.1.0.0
            # 7  0x00007fffe9d46d5b in BIO_free () from /lib/libcrypto.so.1.0.0
            # 8  0x00007fffe9f30a96 in bio_free (bio=0x5555556f3200) at SWIG/_m2crypto_wrap.c:5008
            # 9  0x00007fffe9f30b1e in _wrap_bio_free (self=<optimized out>, args=<optimized out>) at SWIG/_m2crypto_wrap.c
            #
            # We unfortunately have no way to force that order, and there is a risk of deadlock
            # when running in a multi threaded environment like the agents::
            #
            # Thread A opens socket, gets FD = 111
            # Thread A works on it
            # Thread A closes FD 111 (underlying socket.close())
            # Thread B opens socket, gets FD = 111
            # Thread A calls read on FD=111 from ssl3_shutdown
            #
            # This is illustrated on the strace below::
            #
            # 26461 14:25:15.266692 write(111<TCPv6:[[<srcAddressV6>]:42688->[<dstAddressV6>]:9140]>,
            #                            "blabla", 37 <unfinished ...>
            # 26464 14:25:15.266857 <... connect resumed>) = 0 <0.000195>
            # 26464 14:25:15.267023 getsockname(120<UDP:[<srcAddress>:44252->188.185.84.86:9140]>, <unfinished ...>
            # 26461 14:25:15.267176 <... write resumed>) = 37 <0.000453>
            # 26464 14:25:15.267425 <... getsockname resumed>{sa_family=AF_INET, sin_port=htons(44252),
            #                        sin_addr=inet_addr("<srcAddress>")}, [28->16]) = 0 <0.000292>
            # 26461 14:25:15.267466 close(111<TCPv6:[[<srcAddressV6>]:42688->[<dstAddressV6>]:9140]> <unfinished ...>
            # 26464 14:25:15.267637 close(120<UDP:[<srcAddress>:44252->188.185.84.86:9140]> <unfinished ...>
            # 26464 14:25:15.267738 <... close resumed>) = 0 <0.000086>
            # 26461 14:25:15.267768 <... close resumed>) = 0 <0.000285>
            # 26464 14:25:15.267827 socket(AF_INET6, SOCK_DGRAM|SOCK_CLOEXEC, IPPROTO_IP <unfinished ...>
            # 26461 14:25:15.267888 futex(0x21f8620, FUTEX_WAKE_PRIVATE, 1 <unfinished ...>
            # 26464 14:25:15.267976 <... socket resumed>) = 111<UDPv6:[1207802822]> <0.000138>
            # 26461 14:25:15.268092 <... futex resumed>) = 1 <0.000196>
            # 26464 14:25:15.268195 connect(111<UDPv6:[1207802822]>,
            #                      {sa_family=AF_INET6, sin6_port=htons(9140),
            #                       inet_pton(AF_INET6, "<dstAddressV6>", &sin6_addr),
            #                      sin6_flowinfo=htonl(0), sin6_scope_id=0
            #                      }, 28 <unfinished ...>
            # 26461 14:25:15.268294 read(111<UDPv6:[[<srcAddressV6>]:42480->[<dstAddressV6>]:9140]>, <unfinished ...>
            # 26464 14:25:15.268503 <... connect resumed>) = 0 <0.000217>
            # 26464 14:25:15.268673 getsockname(111<UDPv6:[[<srcAddressV6>]:42480->[<dstAddressV6>]:9140]>, <unfinished ...>
            # 26464 14:25:15.268862 <... getsockname resumed>{sa_family=AF_INET6, sin6_port=htons(42480),
            #                        inet_pton(AF_INET6, "<srcAddressV6>", &sin6_addr), sin6_flowinfo=htonl(0), sin6_scope_id=
            # 0}, [28]) = 0 <0.000168>
            # 26464 14:25:15.269048
            # close(111<UDPv6:[[<srcAddressV6>]:42480->[<dstAddressV6>]:9140]>
            # <unfinished ...>
            #
            #
            # Update 16.07.20:
            # M2Crypto 0.36 contains the bug fix https://gitlab.com/m2crypto/m2crypto/-/merge_requests/247
            # that allows proper closing. So manual closing of the underlying socket should not be needed anymore

            # Update 16.07.20
            # I add this shutdown call without being 100% sure
            # it solves some hanging connections issues, but it seems
            # to work. it does not appear in any M2Crypto doc, but comparing
            # some internals of M2Crypto and official python SSL library,
            # it seems to make sense
            self.oSocket.shutdown(socket.SHUT_RDWR)

            # Update 16.07.20
            # With freeBio=True, we force the
            # closing of the socket before the GC runs
            self.oSocket.close(freeBio=True)
            # underlyingSocket = self.oSocket.socket
            self.oSocket = None
            # underlyingSocket.close()
        return S_OK()

    def renewServerContext(self):
        # pylint: disable=line-too-long
        """Renews the server context.
        This reloads the certificates and re-initialises the SSL context.

        NOTE: Chris 15.05.20
        I noticed python segfault on a regular time interval. The stack trace always looks like that::

          #0  0x00007fdb5bbe2388 in ?? () from /opt/dirac/pro/diracos/usr/lib64/python2.7/lib-dynload/../../libcrypto.so.10
          #1  0x00007fdb5bbd8742 in X509_STORE_load_locations () from /opt/dirac/pro/diracos/usr/lib64/python2.7/lib-dynload/../../libcrypto.so.10
          #2  0x00007fdb57edcc9d in _wrap_ssl_ctx_load_verify_locations (self=<optimized out>, args=<optimized out>) at SWIG/_m2crypto_wrap.c:20602
          #3  0x00007fdb644ec484 in PyEval_EvalFrameEx () from /opt/dirac/versions/v10r0_1587978031/diracos/usr/bin/../lib64/libpython2.7.so.1.0

        I could not find anything fundamentaly wrong, and the context renewal is the only place I could think of.

        GSI based SSLTransport did the following: renew the context, and renew the Connection object using the same raw socket
        This still seems very fishy to me though, especially that the ServiceReactor still has the old object in self.__listeningConnections[svcName]['socket']]

        Here, we were are refreshing the CA store. What was missing was the call to the parent class, thus entering some sort of infinite loop.
        The parent's call seems to have fixed it.
        """  # noqa # pylint: disable=line-too-long
        if not self.serverMode():
            raise RuntimeError("SSLTransport is in client mode.")
        super(SSLTransport, self).renewServerContext()
        self.__ctx = getM2SSLContext(self.__ctx, **self.__kwargs)

        return S_OK()

    def handshake_singleStep(self):
        """Used to perform SSL handshakes.
        These are now done automatically.
        """
        # This isn't used any more, the handshake is done inside the M2Crypto library
        return S_OK()

    def handshake_multipleSteps(self):
        """Perform SSL handshakes.
        This has to be called after the connection was accepted (acceptConnection_multipleSteps)

        The remote credentials are gathered here
        """
        try:
            # M2Crypto does not provide public method to
            # accept and handshake in two steps.
            # So we have to do it manually
            # The following lines are basically a copy/paste
            # of the end of SSL.Connection.accept method
            self.oSocket.setup_ssl()
            self.oSocket.set_accept_state()
            self.oSocket.accept_ssl()
            check = getattr(self.oSocket, "postConnectionCheck", self.oSocket.serverPostConnectionCheck)
            if check is not None:
                if not check(self.oSocket.get_peer_cert(), self.oSocket.addr[0]):
                    raise SSL.Checker.SSLVerificationError("post connection check failed")

            self.peerCredentials = getM2PeerInfo(self.oSocket)

            # Now that the handshake has been performed on the server
            # we can set the timeout for the RPC operations.
            # In practice, since we are on the server side, the
            # timeout we set here represents the timeout for receiving
            # the arguments and sending back the response. This should
            # in principle be reasonably quick, but just to be sure
            # we can set it to the DEFAULT_RPC_TIMEOUT
            self.oSocket.settimeout(DEFAULT_RPC_TIMEOUT)

            return S_OK()
        except (socket.error, SSL.SSLError, SSLVerificationError) as e:
            return S_ERROR("Error in handhsake: %s %s" % (e, repr(e)))

    def setClientSocket_singleStep(self, oSocket):
        """Set the inner socket (i.e. SSL.Connection object) of this instance
        to the value of oSocket.
        We also gather the remote peer credentials
        This method is intended to be used to create client connection objects
        from a server and should be considered to be an internal function.

        :param oSocket: client socket SSL.Connection object

        """

        # TODO: The calling method (ServiceReactor.__acceptIncomingConnection) expects
        # socket.error to be thrown in case of issue. Maybe we should catch the M2Crypto
        # errors here and raise socket.error instead

        self.oSocket = oSocket
        self.remoteAddress = self.oSocket.getpeername()
        self.peerCredentials = getM2PeerInfo(self.oSocket)

    def setClientSocket_multipleSteps(self, oSocket):
        """Set the inner socket (i.e. SSL.Connection object) of this instance
        to the value of oSocket.
        This method is intended to be used to create client connection objects
        from a server and should be considered to be an internal function.

        :param oSocket: client socket SSL.Connection object

        """
        # warning: do NOT catch socket.error here, because for who knows what reason
        # exceptions are actually properly used for once, and the calling method
        # relies on it (ServiceReactor.__acceptIncomingConnection)
        self.oSocket = oSocket
        self.remoteAddress = self.oSocket.getpeername()

    def acceptConnection_multipleSteps(self):
        """Accept a new client, returns a new SSLTransport object representing
        the client connection.

        The connection is accepted, but no SSL handshake is performed

        :returns: S_OK(SSLTransport object)
        """
        # M2Crypto does not provide public method to
        # accept and handshake in two steps.
        # So we have to do it manually
        # The following lines are basically a copy/paste
        # of the begining of SSL.Connection.accept method
        # with added options and timeout
        try:
            sock, addr = self.oSocket.socket.accept()
            oClient = SSL.Connection(self.oSocket.ctx, sock)

            # Set the keep alive to true. This keepalive will ensure that we
            # detect remote peer crashing or network interruption
            # Note that this is ineffective if we are in the middle of blocking
            # operations.
            oClient.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, True)

            # I am adding here for reference the code that would allow to change
            # the keepalive settings, although we should be fine with the default
            # (connection would be closed after 7200 + 9 * 75 ~= 2h and 10mn )

            # Duration between two keepalive probes, in seconds
            # oClient.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 75)
            # Number of consecutive bad probes to cut the connection
            # oClient.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 9)
            # Delay before the keepalive starts ticking, in seconds
            # oClient.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 7200)

            # Here we set the timeout server side.
            # We first set the connection timeout, which will
            # be effective for TLS handshake
            oClient.settimeout(DEFAULT_CONNECTION_TIMEOUT)

            oClient.addr = addr
            oClientTrans = SSLTransport(self.stServerAddress, ctx=self.__ctx)
            oClientTrans.setClientSocket(oClient)
            return S_OK(oClientTrans)
        except (socket.error, SSL.SSLError, SSLVerificationError) as e:
            return S_ERROR("Error in acceptConnection: %s %s" % (e, repr(e)))

    def acceptConnection_singleStep(self):
        """Accept a new client, returns a new SSLTransport object representing
        the client connection.

        The SSL handshake is performed here.

        :returns: S_OK(SSLTransport object)
        """
        try:
            oClient, _ = self.oSocket.accept()
            oClientTrans = SSLTransport(self.stServerAddress, ctx=self.__ctx)
            oClientTrans.setClientSocket(oClient)
            return S_OK(oClientTrans)
        except (socket.error, SSL.SSLError, SSLVerificationError) as e:
            return S_ERROR("Error in acceptConnection: %s %s" % (e, repr(e)))

    # Depending on the DIRAC_M2CRYPTO_SPLIT_HANDSHAKE we either do the
    # handshake separately or not
    if os.getenv("DIRAC_M2CRYPTO_SPLIT_HANDSHAKE", "Yes").lower() in ("yes", "true"):
        acceptConnection = acceptConnection_multipleSteps
        handshake = handshake_multipleSteps
        setClientSocket = setClientSocket_multipleSteps
    else:
        acceptConnection = acceptConnection_singleStep
        handshake = handshake_singleStep
        setClientSocket = setClientSocket_singleStep

    def _read(self, bufSize=4096, skipReadyCheck=False):
        """Read bufSize bytes from the buffer.

        :param bufSize: size of the buffer to read
        :param skipReadyCheck: ignored.


        :returns: S_OK(number of byte read)
        """
        try:
            read = self.oSocket.read(bufSize)
            return S_OK(read)
        except (socket.error, SSL.SSLError, SSLVerificationError) as e:
            return S_ERROR("Error in _read: %s %s" % (e, repr(e)))

    def isLocked(self):
        """Returns if this instance is locked.
        Always returns false.

        :returns: False
        """
        return self.__locked

    def _write(self, buf):
        """Write all bytes contained within iterable "buf" to the
        connected peer.

        :param buf: iterable buffer

        :returns: S_OK(number of bytes written)
        """
        try:
            # If the client application has abruptly terminated
            # the connection will be in CLOSE_WAIT on the server side.
            # And when the server side will anyway try to send back its data.
            # The first call to this _write method will succeed,
            # and we will never know that the connection was broken.
            # However, the client would answer with an RST packet.
            # And writting on a socket that received an RST packet
            # triggers a SIGPIPE.
            # In practice, this means that if the server replies to a
            # dead client with less that 16384 bytes (see),
            # we will never notice that we sent the answer to the vacuum.
            # And don't look for a fix, there just isn't.
            wrote = self.oSocket.write(buf)
            return S_OK(wrote)
        except (socket.error, SSL.SSLError, SSLVerificationError) as e:
            return S_ERROR("Error in _write: %s %s" % (e, repr(e)))



"""Utilities to help Computing Element Queues manipulation
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import hashlib

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import fromChar
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getDIRACPlatform
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Resources.Computing.ComputingElementFactory import ComputingElementFactory

__RCSID__ = "$Id$"


def getQueuesResolved(
    siteDict, queueCECache, gridEnv=None, setup=None, workingDir="", checkPlatform=False, instantiateCEs=False
):
    """Get the list of relevant CEs (what is in siteDict) and their descriptions.
    The main goal of this method is to return a dictionary of queues
    """
    queueDict = {}
    ceFactory = ComputingElementFactory()

    for site in siteDict:
        for ce in siteDict[site]:
            ceDict = siteDict[site][ce]
            pilotRunDirectory = ceDict.get("PilotRunDirectory", "")
            # ceMaxRAM = ceDict.get('MaxRAM', None)
            qDict = ceDict.pop("Queues")
            for queue in qDict:
                queueName = "%s_%s" % (ce, queue)
                queueDict[queueName] = {}
                queueDict[queueName]["ParametersDict"] = qDict[queue]
                queueDict[queueName]["ParametersDict"]["Queue"] = queue
                queueDict[queueName]["ParametersDict"]["GridCE"] = ce
                queueDict[queueName]["ParametersDict"]["Site"] = site
                queueDict[queueName]["ParametersDict"]["GridEnv"] = gridEnv
                queueDict[queueName]["ParametersDict"]["Setup"] = setup

                # Evaluate the CPU limit of the queue according to the Glue convention
                computeQueueCPULimit(queueDict[queueName]["ParametersDict"])

                # Tags & RequiredTags defined on the Queue level and on the CE level are concatenated
                # This also converts them from a string to a list if required.
                resolveTags(ceDict, queueDict[queueName]["ParametersDict"])

                # Some parameters can be defined on the CE level and are inherited by all Queues
                setAdditionalParams(ceDict, queueDict[queueName]["ParametersDict"])

                if pilotRunDirectory:
                    queueDict[queueName]["ParametersDict"]["JobExecDir"] = pilotRunDirectory
                qwDir = os.path.join(workingDir, queue)
                mkDir(qwDir)
                queueDict[queueName]["ParametersDict"]["WorkingDirectory"] = qwDir
                ceQueueDict = dict(ceDict)
                ceQueueDict.update(queueDict[queueName]["ParametersDict"])

                if instantiateCEs:
                    # Generate the CE object for the queue or pick the already existing one
                    # if the queue definition did not change
                    queueHash = generateQueueHash(ceQueueDict)
                    if queueName in queueCECache and queueCECache[queueName]["Hash"] == queueHash:
                        queueCE = queueCECache[queueName]["CE"]
                    else:
                        result = ceFactory.getCE(ceName=ce, ceType=ceDict["CEType"], ceParametersDict=ceQueueDict)
                        if not result["OK"]:
                            continue
                        queueCECache.setdefault(queueName, {})
                        queueCECache[queueName]["Hash"] = queueHash
                        queueCECache[queueName]["CE"] = result["Value"]
                        queueCE = queueCECache[queueName]["CE"]

                    queueDict[queueName]["ParametersDict"].update(queueCE.ceParameters)
                    queueDict[queueName]["CE"] = queueCE
                    result = queueDict[queueName]["CE"].isValid()
                    if not result["OK"]:
                        continue

                queueDict[queueName]["CEName"] = ce
                queueDict[queueName]["CEType"] = ceDict["CEType"]
                queueDict[queueName]["Site"] = site
                queueDict[queueName]["QueueName"] = queue
                queueDict[queueName]["QueryCEFlag"] = ceDict.get("QueryCEFlag", "false")

                if checkPlatform:
                    setPlatform(ceDict, queueDict[queueName]["ParametersDict"])

                bundleProxy = queueDict[queueName]["ParametersDict"].get("BundleProxy", ceDict.get("BundleProxy"))
                if bundleProxy and bundleProxy.lower() in ["true", "yes", "1"]:
                    queueDict[queueName]["BundleProxy"] = True

    return S_OK(queueDict)


def computeQueueCPULimit(queueDict):
    """Evaluate the CPU limit of the queue according to the Glue convention"""
    if "maxCPUTime" in queueDict and "SI00" in queueDict:
        maxCPUTime = float(queueDict["maxCPUTime"])
        # For some sites there are crazy values in the CS
        maxCPUTime = max(maxCPUTime, 0)
        maxCPUTime = min(maxCPUTime, 86400 * 12.5)
        si00 = float(queueDict["SI00"])
        queueCPUTime = 60 / 250 * maxCPUTime * si00
        queueDict["CPUTime"] = int(queueCPUTime)


def resolveTags(ceDict, queueDict):
    """Tags & RequiredTags defined on the Queue level and on the CE level are concatenated.
    This also converts them from a string to a list if required.
    """
    for tagFieldName in ("Tag", "RequiredTag"):
        ceTags = ceDict.get(tagFieldName, [])
        if isinstance(ceTags, six.string_types):
            ceTags = fromChar(ceTags)
        queueTags = queueDict.get(tagFieldName, [])
        if isinstance(queueTags, six.string_types):
            queueTags = fromChar(queueTags)
        queueDict[tagFieldName] = list(set(ceTags) | set(queueTags))


def setPlatform(ceDict, queueDict):
    """Set platform according to CE parameters if not defined"""
    platform = queueDict.get("Platform", ceDict.get("Platform", ""))
    if not platform and "OS" in ceDict:
        architecture = ceDict.get("architecture", "x86_64")
        platform = "_".join([architecture, ceDict["OS"]])

    if "Platform" not in queueDict and platform:
        result = getDIRACPlatform(platform)
        if result["OK"]:
            queueDict["Platform"] = result["Value"][0]
        else:
            queueDict["Platform"] = platform


def setAdditionalParams(ceDict, queueDict):
    """Some parameters can be defined on the CE level and are inherited by all Queues"""
    for parameter in ["MaxRAM", "NumberOfProcessors", "WholeNode"]:
        queueParameter = queueDict.get(parameter, ceDict.get(parameter))
        if queueParameter:
            queueDict[parameter] = queueParameter


def generateQueueHash(queueDict):
    """Generate a hash of the queue description"""
    myMD5 = hashlib.md5()
    myMD5.update(str(queueDict).encode())
    hexstring = myMD5.hexdigest()
    return hexstring


def matchQueue(jobJDL, queueDict, fullMatch=False):
    """
    Match the job description to the queue definition

    :param str job: JDL job description
    :param bool fullMatch: test matching on all the criteria
    :param dict queueDict: queue parameters dictionary

    :return: S_OK/S_ERROR, Value - result of matching, S_OK if matched or
             S_ERROR with the reason for no match
    """

    # Check the job description validity
    job = ClassAd(jobJDL)
    if not job.isOK():
        return S_ERROR("Invalid job description")

    noMatchReasons = []

    # Check job requirements to resource
    # 1. CPUTime
    cpuTime = job.getAttributeInt("CPUTime")
    if not cpuTime:
        cpuTime = 84600
    if cpuTime > int(queueDict.get("CPUTime", 0)):
        noMatchReasons.append("Job CPUTime requirement not satisfied")
        if not fullMatch:
            return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 2. Multi-value match requirements
    for parameter in ["Site", "GridCE", "Platform", "SubmitPool", "JobType"]:
        if parameter in queueDict:
            valueSet = set(job.getListFromExpression(parameter))
            if not valueSet:
                valueSet = set(job.getListFromExpression("%ss" % parameter))
            queueSet = set(fromChar(queueDict[parameter]))
            if valueSet and queueSet and not valueSet.intersection(queueSet):
                valueToPrint = ",".join(valueSet)
                if len(valueToPrint) > 20:
                    valueToPrint = "%s..." % valueToPrint[:20]
                noMatchReasons.append("Job %s %s requirement not satisfied" % (parameter, valueToPrint))
                if not fullMatch:
                    return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 3. Banned multi-value match requirements
    for par in ["Site", "GridCE", "Platform", "JobType"]:
        parameter = "Banned%s" % par
        if par in queueDict:
            valueSet = set(job.getListFromExpression(parameter))
            if not valueSet:
                valueSet = set(job.getListFromExpression("%ss" % parameter))
            queueSet = set(fromChar(queueDict[par]))
            if valueSet and queueSet and valueSet.issubset(queueSet):
                valueToPrint = ",".join(valueSet)
                if len(valueToPrint) > 20:
                    valueToPrint = "%s..." % valueToPrint[:20]
                noMatchReasons.append("Job %s %s requirement not satisfied" % (parameter, valueToPrint))
                if not fullMatch:
                    return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 4. Tags
    tags = set(job.getListFromExpression("Tag"))
    nProc = job.getAttributeInt("NumberOfProcessors")
    if nProc and nProc > 1:
        tags.add("MultiProcessor")
    wholeNode = job.getAttributeString("WholeNode")
    if wholeNode:
        tags.add("WholeNode")
    queueTags = set(queueDict.get("Tag", []))
    if not tags.issubset(queueTags):
        noMatchReasons.append("Job Tag %s not satisfied" % ",".join(tags))
        if not fullMatch:
            return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 4. MultiProcessor requirements
    if nProc and nProc > int(queueDict.get("NumberOfProcessors", 1)):
        noMatchReasons.append("Job NumberOfProcessors %d requirement not satisfied" % nProc)
        if not fullMatch:
            return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 5. RAM
    ram = job.getAttributeInt("RAM")
    # If MaxRAM is not specified in the queue description, assume 2GB
    if ram and ram > int(queueDict.get("MaxRAM", 2048) / 1024):
        noMatchReasons.append("Job RAM %d requirement not satisfied" % ram)
        if not fullMatch:
            return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # Check resource requirements to job
    # 1. OwnerGroup - rare case but still
    if "OwnerGroup" in queueDict:
        result = getProxyInfo(disableVOMS=True)
        if not result["OK"]:
            return S_ERROR("No valid proxy available")
        ownerGroup = result["Value"]["group"]
        if ownerGroup != queueDict["OwnerGroup"]:
            noMatchReasons.append("Resource OwnerGroup %s requirement not satisfied" % queueDict["OwnerGroup"])
            if not fullMatch:
                return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 2. Required tags
    requiredTags = set(queueDict.get("RequiredTags", []))
    if not requiredTags.issubset(tags):
        noMatchReasons.append("Resource RequiredTags %s not satisfied" % ",".join(requiredTags))
        if not fullMatch:
            return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    # 3. RunningLimit
    site = queueDict["Site"]
    ce = queueDict.get("GridCE")
    opsHelper = Operations()
    result = opsHelper.getSections("JobScheduling/RunningLimit")
    if result["OK"] and site in result["Value"]:
        result = opsHelper.getSections("JobScheduling/RunningLimit/%s" % site)
        if result["OK"]:
            for parameter in result["Value"]:
                value = job.getAttributeString(parameter)
                if (
                    value
                    and (
                        opsHelper.getValue("JobScheduling/RunningLimit/%s/%s/%s" % (site, parameter, value), 1)
                        or opsHelper.getValue(
                            "JobScheduling/RunningLimit/%s/CEs/%s/%s/%s" % (site, ce, parameter, value), 1
                        )
                    )
                    == 0
                ):
                    noMatchReasons.append("Resource operational %s requirement not satisfied" % parameter)
                    if not fullMatch:
                        return S_OK({"Match": False, "Reason": noMatchReasons[0]})

    return S_OK({"Match": not bool(noMatchReasons), "Reason": noMatchReasons})



"""
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from diraccfg import CFG

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.JDL import loadJDLAsCFG, dumpCFGAsJDL


class JobManifest(object):
    def __init__(self, manifest=""):
        self.__manifest = CFG()
        self.__dirty = False
        self.__ops = False
        if manifest:
            result = self.load(manifest)
            if not result["OK"]:
                raise Exception(result["Message"])

    def isDirty(self):
        return self.__dirty

    def setDirty(self):
        self.__dirty = True

    def clearDirty(self):
        self.__dirty = False

    def load(self, dataString):
        """
        Auto discover format type based on [ .. ] of JDL
        """
        dataString = dataString.strip()
        if dataString[0] == "[" and dataString[-1] == "]":
            return self.loadJDL(dataString)
        else:
            return self.loadCFG(dataString)

    def loadJDL(self, jdlString):
        """
        Load job manifest from JDL format
        """
        result = loadJDLAsCFG(jdlString.strip())
        if not result["OK"]:
            self.__manifest = CFG()
            return result
        self.__manifest = result["Value"][0]
        return S_OK()

    def loadCFG(self, cfgString):
        """
        Load job manifest from CFG format
        """
        try:
            self.__manifest.loadFromBuffer(cfgString)
        except Exception as e:
            return S_ERROR("Can't load manifest from cfg: %s" % str(e))
        return S_OK()

    def dumpAsCFG(self):
        return str(self.__manifest)

    def getAsCFG(self):
        return self.__manifest.clone()

    def dumpAsJDL(self):
        return dumpCFGAsJDL(self.__manifest)

    def __getCSValue(self, varName, defaultVal=None):
        if not self.__ops:
            self.__ops = Operations(group=self.__manifest["OwnerGroup"], setup=self.__manifest["DIRACSetup"])
        if varName[0] != "/":
            varName = "JobDescription/%s" % varName
        return self.__ops.getValue(varName, defaultVal)

    def __checkNumericalVar(self, varName, defaultVal, minVal, maxVal):
        """
        Check a numerical var
        """
        initialVal = False
        if varName not in self.__manifest:
            varValue = self.__getCSValue("Default%s" % varName, defaultVal)
        else:
            varValue = self.__manifest[varName]
            initialVal = varValue
        try:
            varValue = int(varValue)
        except ValueError:
            return S_ERROR("%s must be a number" % varName)
        minVal = self.__getCSValue("Min%s" % varName, minVal)
        maxVal = self.__getCSValue("Max%s" % varName, maxVal)
        varValue = max(minVal, min(varValue, maxVal))
        if initialVal != varValue:
            self.__manifest.setOption(varName, varValue)
        return S_OK(varValue)

    def __checkChoiceVar(self, varName, defaultVal, choices):
        """
        Check a choice var
        """
        initialVal = False
        if varName not in self.__manifest:
            varValue = self.__getCSValue("Default%s" % varName, defaultVal)
        else:
            varValue = self.__manifest[varName]
            initialVal = varValue
        if varValue not in self.__getCSValue("Choices%s" % varName, choices):
            return S_ERROR("%s is not a valid value for %s" % (varValue, varName))
        if initialVal != varValue:
            self.__manifest.setOption(varName, varValue)
        return S_OK(varValue)

    def __checkMultiChoice(self, varName, choices):
        """
        Check a multi choice var
        """
        initialVal = False
        if varName not in self.__manifest:
            return S_OK()
        else:
            varValue = self.__manifest[varName]
            initialVal = varValue
        choices = self.__getCSValue("Choices%s" % varName, choices)
        for v in List.fromChar(varValue):
            if v not in choices:
                return S_ERROR("%s is not a valid value for %s" % (v, varName))
        if initialVal != varValue:
            self.__manifest.setOption(varName, varValue)
        return S_OK(varValue)

    def __checkMaxInputData(self, maxNumber):
        """
        Check Maximum Number of Input Data files allowed
        """
        varName = "InputData"
        if varName not in self.__manifest:
            return S_OK()
        varValue = self.__manifest[varName]
        if len(List.fromChar(varValue)) > maxNumber:
            return S_ERROR(
                "Number of Input Data Files (%s) greater than current limit: %s"
                % (len(List.fromChar(varValue)), maxNumber)
            )
        return S_OK()

    def __contains__(self, key):
        """Check if the manifest has the required key"""
        return key in self.__manifest

    def setOptionsFromDict(self, varDict):
        for k in sorted(varDict):
            self.setOption(k, varDict[k])

    def check(self):
        """
        Check that the manifest is OK
        """
        for k in ["OwnerName", "OwnerDN", "OwnerGroup", "DIRACSetup"]:
            if k not in self.__manifest:
                return S_ERROR("Missing var %s in manifest" % k)

        # Check CPUTime
        result = self.__checkNumericalVar("CPUTime", 86400, 100, 500000)
        if not result["OK"]:
            return result

        result = self.__checkNumericalVar("Priority", 1, 0, 10)
        if not result["OK"]:
            return result

        maxInputData = Operations().getValue("JobDescription/MaxInputData", 500)
        result = self.__checkMaxInputData(maxInputData)
        if not result["OK"]:
            return result

        operation = Operations(group=self.__manifest["OwnerGroup"])
        allowedJobTypes = operation.getValue("JobDescription/AllowedJobTypes", ["User", "Test", "Hospital"])
        transformationTypes = operation.getValue("Transformations/DataProcessing", [])
        result = self.__checkMultiChoice("JobType", allowedJobTypes + transformationTypes)
        if not result["OK"]:
            return result
        return S_OK()

    def createSection(self, secName, contents=False):
        if secName not in self.__manifest:
            if contents and not isinstance(contents, CFG):
                return S_ERROR("Contents for section %s is not a cfg object" % secName)
            self.__dirty = True
            return S_OK(self.__manifest.createNewSection(secName, contents=contents))
        return S_ERROR("Section %s already exists" % secName)

    def getSection(self, secName):
        self.__dirty = True
        if secName not in self.__manifest:
            return S_ERROR("%s does not exist" % secName)
        sec = self.__manifest[secName]
        if not sec:
            return S_ERROR("%s section empty" % secName)
        return S_OK(sec)

    def setSectionContents(self, secName, contents):
        if contents and not isinstance(contents, CFG):
            return S_ERROR("Contents for section %s is not a cfg object" % secName)
        self.__dirty = True
        if secName in self.__manifest:
            self.__manifest[secName].reset()
            self.__manifest[secName].mergeWith(contents)
        else:
            self.__manifest.createNewSection(secName, contents=contents)

    def setOption(self, varName, varValue):
        """
        Set a var in job manifest
        """
        self.__dirty = True
        levels = List.fromChar(varName, "/")
        cfg = self.__manifest
        for l in levels[:-1]:
            if l not in cfg:
                cfg.createNewSection(l)
            cfg = cfg[l]
        cfg.setOption(levels[-1], varValue)

    def remove(self, opName):
        levels = List.fromChar(opName, "/")
        cfg = self.__manifest
        for l in levels[:-1]:
            if l not in cfg:
                return S_ERROR("%s does not exist" % opName)
            cfg = cfg[l]
        if cfg.deleteKey(levels[-1]):
            self.__dirty = True
            return S_OK()
        return S_ERROR("%s does not exist" % opName)

    def getOption(self, varName, defaultValue=None):
        """
        Get a variable from the job manifest
        """
        cfg = self.__manifest
        return cfg.getOption(varName, defaultValue)

    def getOptionList(self, section=""):
        """
        Get a list of variables in a section of the job manifest
        """
        cfg = self.__manifest.getRecursive(section)
        if not cfg or "value" not in cfg:
            return []
        cfg = cfg["value"]
        return cfg.listOptions()

    def isOption(self, opName):
        """
        Check if it is a valid option
        """
        return self.__manifest.isOption(opName)

    def getSectionList(self, section=""):
        """
        Get a list of sections in the job manifest
        """
        cfg = self.__manifest.getRecursive(section)
        if not cfg or "value" not in cfg:
            return []
        cfg = cfg["value"]
        return cfg.listSections()



""" JobCommand

  The JobCommand class is a command class to know about present jobs efficiency

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.ResourceStatusSystem.Command.Command import Command
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.WorkloadManagementSystem.Client.WMSAdministratorClient import WMSAdministratorClient


class JobCommand(Command):
    """
    Job "master" Command.
    """

    def __init__(self, args=None, clients=None):

        super(JobCommand, self).__init__(args, clients)

        if "WMSAdministrator" in self.apis:
            self.wmsAdmin = self.apis["WMSAdministrator"]
        else:
            self.wmsAdmin = WMSAdministratorClient()

        if "ResourceManagementClient" in self.apis:
            self.rmClient = self.apis["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def _storeCommand(self, result):
        """
        Stores the results of doNew method on the database.
        """

        for jobDict in result:

            resQuery = self.rmClient.addOrModifyJobCache(
                jobDict["Site"], jobDict["MaskStatus"], jobDict["Efficiency"], jobDict["Status"]
            )
            if not resQuery["OK"]:
                return resQuery
        return S_OK()

    def _prepareCommand(self):
        """
        JobCommand requires one arguments:
        - name : <str>
        """

        if "name" not in self.args:
            return S_ERROR('"name" not found in self.args')
        name = self.args["name"]

        return S_OK(name)

    def doNew(self, masterParams=None):
        """
        Gets the parameters to run, either from the master method or from its
        own arguments.

        It contacts the WMSAdministrator with a list of site names, or a single
        site.

        If there are jobs, are recorded and then returned.
        """

        if masterParams is not None:
            name = masterParams
        else:
            params = self._prepareCommand()
            if not params["OK"]:
                return params
            name = params["Value"]

        # selectDict, sortList, startItem, maxItems
        # Returns statistics of Last day !
        results = self.wmsAdmin.getSiteSummaryWeb({"Site": name}, [], 0, 0)
        if not results["OK"]:
            return results
        results = results["Value"]

        if "ParameterNames" not in results:
            return S_ERROR('Wrong result dictionary, missing "ParameterNames"')
        params = results["ParameterNames"]

        if "Records" not in results:
            return S_ERROR('Wrong formed result dictionary, missing "Records"')
        records = results["Records"]

        uniformResult = []

        for record in records:

            # This returns a dictionary with the following keys
            # 'Site', 'GridType', 'Country', 'Tier', 'MaskStatus', 'Received',
            # 'Checking', 'Staging', 'Waiting', 'Matched', 'Running', 'Stalled',
            # 'Done', 'Completed', 'Failed', 'Efficiency', 'Status'
            jobDict = dict(zip(params, record))

            # We cast efficiency to a float
            jobDict["Efficiency"] = float(jobDict["Efficiency"])

            uniformResult.append(jobDict)

        storeRes = self._storeCommand(uniformResult)
        if not storeRes["OK"]:
            return storeRes

        return S_OK(uniformResult)

    def doCache(self):
        """
        Method that reads the cache table and tries to read from it. It will
        return a list of dictionaries if there are results.
        """

        params = self._prepareCommand()
        if not params["OK"]:
            return params
        name = params["Value"]

        result = self.rmClient.selectJobCache(name)
        if result["OK"]:
            result = S_OK([dict(zip(result["Columns"], res)) for res in result["Value"]])

        return result

    def doMaster(self):
        """
        Master method.

        Gets all sites and calls doNew method.
        """

        siteNames = getSites()
        if not siteNames["OK"]:
            return siteNames
        siteNames = siteNames["Value"]

        jobsResults = self.doNew(siteNames)
        if not jobsResults["OK"]:
            self.metrics["failed"].append(jobsResults["Message"])

        return S_OK(self.metrics)


class JobsWMSCommand(Command):
    def __init__(self, args=None, clients=None):

        super(JobsWMSCommand, self).__init__(args, clients)

        if "WMSAdministrator" in self.apis:
            self.wmsAdmin = self.apis["WMSAdministrator"]
        else:
            self.wmsAdmin = WMSAdministratorClient()

    def doCommand(self):
        """
        Returns simple jobs efficiency

        :param args:
           - args[0]: string: should be a ValidElement

           - args[1]: string should be the name of the ValidElement

        :returns: { 'Result': 'Good'|'Fair'|'Poor'|'Idle'|'Bad' }
        """

        if "siteName" not in self.args:
            return self.returnERROR(S_ERROR("siteName is missing"))
        siteName = self.args["siteName"]

        # If siteName is None, we take all sites
        if siteName is None:
            siteName = getSites()
            if not siteName["OK"]:
                return self.returnERROR(siteName)
            siteName = siteName["Value"]

        results = self.wmsAdmin.getSiteSummaryWeb({"Site": siteName}, [], 0, 500)

        if not results["OK"]:
            return self.returnERROR(results)
        results = results["Value"]

        if "ParameterNames" not in results:
            return self.returnERROR(S_ERROR("Malformed result dictionary"))
        params = results["ParameterNames"]

        if "Records" not in results:
            return self.returnERROR(S_ERROR("Malformed result dictionary"))
        records = results["Records"]

        jobResults = []

        for record in records:

            jobDict = dict(zip(params, record))
            try:
                jobDict["Efficiency"] = float(jobDict["Efficiency"])
            except KeyError as e:
                return self.returnERROR(S_ERROR(e))
            except ValueError as e:
                return self.returnERROR(S_ERROR(e))

            jobResults.append(jobDict)

        return S_OK(jobResults)



""" DIRAC FileCatalog component representing a directory tree with enumerated paths
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access


__RCSID__ = "$Id$"

import six
import os

from DIRAC import S_OK, S_ERROR
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase

MAX_LEVELS = 15


class DirectoryLevelTree(DirectoryTreeBase):
    """Class managing Directory Tree as a simple self-linked structure
    with full directory path stored in each node
    """

    def __init__(self, database=None):
        DirectoryTreeBase.__init__(self, database)
        self.treeTable = "FC_DirectoryLevelTree"

    def getTreeType(self):

        return "Directory"

    def findDir(self, path, connection=False):
        """Find directory ID for the given path"""

        dpath = self.db._escapeString(os.path.normpath(path))
        if not dpath["OK"]:
            return dpath
        dpath = dpath["Value"]
        req = "SELECT DirID,Level from FC_DirectoryLevelTree WHERE DirName=%s" % dpath
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result

        if not result["Value"]:
            return S_OK("")

        res = S_OK(result["Value"][0][0])
        res["Level"] = result["Value"][0][1]
        return res

    def findDirs(self, paths, connection=False):
        """Find DirIDs for the given path list"""
        dpathList = []
        for path in paths:
            dpath = self.db._escapeString(os.path.normpath(path))
            if not dpath["OK"]:
                return dpath
            dpathList.append(dpath["Value"])
        dpaths = ",".join(dpathList)
        req = "SELECT DirName,DirID from FC_DirectoryLevelTree WHERE DirName in (%s)" % dpaths
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        dirDict = {}
        for dirName, dirID in result["Value"]:
            dirDict[dirName] = dirID

        return S_OK(dirDict)

    def removeDir(self, path):
        """Remove directory"""

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            res = S_OK()
            res["DirID"] = 0
            return res

        dirID = result["Value"]
        req = "DELETE FROM FC_DirectoryLevelTree WHERE DirID=%d" % dirID
        result = self.db._update(req)
        result["DirID"] = dirID
        return result

    def __getNumericPath(self, dirID, connection=False):
        """Get the enumerated path of the given directory"""
        epathString = ",".join(["LPATH%d" % (i + 1) for i in range(MAX_LEVELS)])
        req = "SELECT LEVEL,%s FROM FC_DirectoryLevelTree WHERE DirID=%d" % (epathString, dirID)
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        row = result["Value"][0]
        level = row[0]
        epathList = []
        for i in range(level):
            epathList.append(row[i + 1])

        result = S_OK(epathList)
        result["Level"] = level
        return result

    def makeDir(self, path):
        """Create a new directory entry"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        if dirID:
            result = S_OK(dirID)
            result["NewDirectory"] = False
            return result

        dpath = path
        if path == "/":
            level = 0
            elements = []
            parentDirID = 0
        else:
            if path[0] == "/":
                dpath = path[1:]
            elements = dpath.split("/")
            level = len(elements)
            if level > MAX_LEVELS:
                return S_ERROR("Too many directory levels: %d" % level)
            result = self.getParent(path)
            if not result["OK"]:
                return result
            parentDirID = result["Value"]

        epathList = []
        if parentDirID:
            result = self.__getNumericPath(parentDirID)
            if not result["OK"]:
                return result
            epathList = result["Value"]

        names = ["DirName", "Level", "Parent"]
        values = [path, level, parentDirID]
        if path != "/":
            for i in range(1, level, 1):
                names.append("LPATH%d" % i)
                values.append(epathList[i - 1])

        result = self.db._getConnection()
        conn = result["Value"]
        # result = self.db._query("LOCK TABLES FC_DirectoryLevelTree WRITE; ",conn)
        result = self.db.insertFields("FC_DirectoryLevelTree", names, values, conn)
        if not result["OK"]:
            # resUnlock = self.db._query("UNLOCK TABLES;",conn)
            if result["Message"].find("Duplicate") != -1:
                # The directory is already added
                resFind = self.findDir(path)
                if not resFind["OK"]:
                    return resFind
                dirID = resFind["Value"]
                result = S_OK(dirID)
                result["NewDirectory"] = False
                return result
            else:
                return result
        dirID = result["lastRowId"]

        # Update the path number
        if parentDirID:
            #       lPath = "LPATH%d" % (level)
            #       req = " SELECT @tmpvar:=max(%s)+1 FROM FC_DirectoryLevelTree WHERE Parent=%d; " % (lPath,parentDirID)
            #       resultLock = self.db._query("LOCK TABLES FC_DirectoryLevelTree WRITE; ",conn)
            #       result = self.db._query(req,conn)
            #       req = "UPDATE FC_DirectoryLevelTree SET %s=@tmpvar WHERE DirID=%d; " % (lPath,dirID)
            #       result = self.db._update(req,conn)
            #       result = self.db._query("UNLOCK TABLES;",conn)
            lPath = "LPATH%d" % (level)
            req = " SELECT @tmpvar:=max(%s)+1 FROM FC_DirectoryLevelTree WHERE Parent=%d FOR UPDATE; " % (
                lPath,
                parentDirID,
            )
            resultLock = self.db._query("START TRANSACTION; ", conn)
            result = self.db._query(req, conn)
            req = "UPDATE FC_DirectoryLevelTree SET %s=@tmpvar WHERE DirID=%d; " % (lPath, dirID)
            result = self.db._update(req, conn)
            result = self.db._query("COMMIT;", conn)
            if not result["OK"]:
                return result
        else:
            result = self.db._query("ROLLBACK;", conn)

        result = S_OK(dirID)
        result["NewDirectory"] = True
        return result

    def existsDir(self, path):
        """Check the existence of a directory at the specified path"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({"Exists": False})
        return S_OK({"Exists": True, "DirID": result["Value"]})

    def getParent(self, path):
        """Get the parent ID of the given directory"""

        parent_dir = os.path.dirname(path)
        return self.findDir(parent_dir)

    def getParentID(self, dirPathOrID):
        """Get the ID of the parent of a directory specified by ID"""

        dirID = dirPathOrID
        if isinstance(dirPathOrID, six.string_types):
            result = self.findDir(dirPathOrID)
            if not result["OK"]:
                return result
            dirID = result["Value"]

        if dirID == 0:
            return S_ERROR("Root directory ID given")

        req = "SELECT Parent FROM FC_DirectoryLevelTree WHERE DirID=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("No parent found")

        return S_OK(result["Value"][0][0])

    def getDirectoryPath(self, dirID):
        """Get directory name by directory ID"""
        req = "SELECT DirName FROM FC_DirectoryLevelTree WHERE DirID=%d" % int(dirID)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory with id %d not found" % int(dirID))

        return S_OK(result["Value"][0][0])

    def getDirectoryPaths(self, dirIDList):
        """Get directory name by directory ID list"""
        dirs = dirIDList
        if not isinstance(dirIDList, list):
            dirs = [dirIDList]

        if not dirs:
            return S_OK({})

        dirListString = ",".join([str(d) for d in dirs])
        req = "SELECT DirID,DirName FROM FC_DirectoryLevelTree WHERE DirID in ( %s )" % dirListString
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directories not found: %s" % dirListString)

        resultDict = {}
        for row in result["Value"]:
            resultDict[int(row[0])] = row[1]

        return S_OK(resultDict)

    def getDirectoryName(self, dirID):
        """Get directory name by directory ID"""

        result = self.getDirectoryPath(dirID)
        if not result["OK"]:
            return result

        return S_OK(os.path.basename(result["Value"]))

    def getPathIDs(self, path):
        """Get IDs of all the directories in the parent hierarchy for a directory
        specified by its path
        """

        elements = path.split("/")
        pelements = []
        dPath = ""
        for el in elements[1:]:
            dPath += "/" + el
            pelements.append(dPath)
        pelements.append("/")

        pathString = ["'" + p + "'" for p in pelements]
        req = "SELECT DirID FROM FC_DirectoryLevelTree WHERE DirName in (%s) ORDER BY DirID" % ",".join(pathString)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory %s not found" % path)

        return S_OK([x[0] for x in result["Value"]])

    def getPathIDsByID_old(self, dirID):
        """Get IDs of all the directories in the parent hierarchy for a directory
        specified by its ID
        """

        # The method should be rather implemented using enumerated paths

        result = self.getDirectoryPath(dirID)
        if not result["OK"]:
            return result
        dPath = result["Value"]
        return self.getPathIDs(dPath)

    def getPathIDsByID(self, dirID):
        """Get IDs of all the directories in the parent hierarchy for a directory
        specified by its ID
        """
        result = self.__getNumericPath(dirID)
        if not result["OK"]:
            return result
        level = result["Level"]
        if level == 0:
            return S_OK([dirID])
        lpaths = result["Value"]

        lpathSelects = []
        for lev in range(level):
            sel = " AND ".join(["Level=%d" % lev] + ["LPATH%d=%d" % (ll + 1, lpaths[ll]) for ll in range(lev)])
            lpathSelects.append(sel)
        selection = "(" + ") OR (".join(lpathSelects) + ")"
        req = "SELECT Level,DirID from FC_DirectoryLevelTree WHERE %s ORDER BY Level" % selection
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("No result for the path of Directory with ID %d" % dirID)

        return S_OK([x[1] for x in result["Value"]] + [dirID])

    def getChildren(self, path, connection=False):
        """Get child directory IDs for the given directory"""
        if isinstance(path, six.string_types):
            result = self.findDir(path, connection)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Directory does not exist: %s" % path)
            dirID = result["Value"]
        else:
            dirID = path
        req = "SELECT DirID FROM FC_DirectoryLevelTree WHERE Parent=%d" % dirID
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        return S_OK([x[0] for x in result["Value"]])

    def getSubdirectoriesByID(self, dirID, requestString=False, includeParent=False):
        """Get all the subdirectories of the given directory at a given level"""
        req = "SELECT Level FROM FC_DirectoryLevelTree WHERE DirID=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory %d not found" % dirID)
        level = result["Value"][0][0]

        sPaths = []
        if requestString:
            req = "SELECT DirID FROM FC_DirectoryLevelTree"
        else:
            req = "SELECT Level,DirID FROM FC_DirectoryLevelTree"
        if level > 0:
            req += " AS F1"
            for i in range(1, level + 1):
                sPaths.append("LPATH%d" % i)
            pathString = ",".join(sPaths)
            req += " JOIN (SELECT %s FROM FC_DirectoryLevelTree WHERE DirID=%d) AS F2 ON " % (pathString, dirID)
            sPaths = []
            for i in range(1, level + 1):
                sPaths.append("F1.LPATH%d=F2.LPATH%d" % (i, i))
            pString = " AND ".join(sPaths)
            if includeParent:
                req += "%s AND F1.Level >= %d" % (pString, level)
            else:
                req += "%s AND F1.Level > %d" % (pString, level)

        if requestString:
            return S_OK(req)

        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})

        resDict = {}
        for row in result["Value"]:
            resDict[row[1]] = row[0]

        return S_OK(resDict)

    def countSubdirectories(self, dirId, includeParent=True):
        result = self.getSubdirectoriesByID(dirId, requestString=True, includeParent=includeParent)
        if not result["OK"]:
            return result
        reqDir = result["Value"].replace("SELECT DirID FROM", "SELECT count(*) FROM")

        result = self.db._query(reqDir)
        if not result["OK"]:
            return result
        return S_OK(result["Value"][0][0])

    def getAllSubdirectoriesByID(self, dirList):
        """Get IDs of all the subdirectories of directories in a given list"""

        dirs = dirList
        if not isinstance(dirList, list):
            dirs = [dirList]
        resultList = []
        parentList = dirs
        while parentList:
            subResult = []
            dirListString = ",".join([str(d) for d in parentList])
            req = "SELECT DirID from FC_DirectoryLevelTree WHERE Parent in ( %s )" % dirListString
            result = self.db._query(req)
            if not result["OK"]:
                return result
            for row in result["Value"]:
                subResult.append(row[0])
            if subResult:
                resultList += subResult
            parentList = subResult

        return S_OK(resultList)

    def getSubdirectories(self, path):
        """Get subdirectories of the given directory"""

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})

        dirID = result["Value"]
        result = self.getSubdirectoriesByID(dirID)
        return result

    def recoverOrphanDirectories(self, credDict):
        """Recover orphan directories"""
        # Find out orphan directories
        treeTable = "FC_DirectoryLevelTree"
        req = "SELECT DirID,Parent,Level FROM %s WHERE Parent NOT IN ( SELECT DirID from %s )" % (treeTable, treeTable)
        result = self.db._query(req)
        if not result["OK"]:
            return result

        parentDict = {}
        for dirID, parentID, _level in result["Value"]:

            result = self.getDirectoryPath(dirID)
            if not result["OK"]:
                continue
            dirPath = result["Value"]
            parentPath = os.path.dirname(dirPath)
            if not dirPath == "/":
                parentDict.setdefault(parentPath, {})
                parentDict[parentPath].setdefault("DirList", [])
                parentDict[parentPath]["DirList"].append(dirID)
                parentDict[parentPath]["OldParentID"] = parentID

        for parentPath, dirDict in parentDict.items():
            dirIDList = dirDict["DirList"]
            oldParentID = dirDict["OldParentID"]
            result = self.findDir(parentPath)
            if not result["OK"]:
                continue
            if result["Value"]:
                # The parent directory was recreated already
                parentID = result["Value"]
            else:
                # The parent directory was lost
                result = self.makeDirectories(parentPath, credDict)
                if not result["OK"]:
                    continue
                parentID = result["Value"]
                # We have created a new directory but let's keep the old ID
                req = "UPDATE FC_DirectoryLevelTree SET DirID=%s WHERE DirID=%s" % (oldParentID, parentID)
                result = self.db._update(req)
                if not result["OK"]:
                    continue
                req = "UPDATE FC_DirectoryInfo SET DirID=%s WHERE DirID=%s" % (oldParentID, parentID)
                result = self.db._update(req)

                parentID = oldParentID
                # We have to change also the ownership of the new directory to the most likely one
                # which is the owner of the containing directory
                containerPath = os.path.dirname(parentPath)
                result = self.getDirectoryParameters(containerPath)
                if result["OK"]:
                    conDict = result["Value"]
                    uid = conDict["UID"]
                    gid = conDict["GID"]
                    result = self._setDirectoryParameter(parentID, "UID", uid)
                    result = self._setDirectoryParameter(parentID, "GID", gid)

            dirString = ",".join([str(dirID) for dirID in dirIDList])
            req = "UPDATE FC_DirectoryLevelTree SET Parent=%s WHERE DirID IN (%s)" % (parentID, dirString)
            result = self.db._update(req)
            if not result["OK"]:
                continue

            connection = self._getConnection()
            result = self.db._query("LOCK TABLES FC_DirectoryLevelTree WRITE", connection)
            if not result["OK"]:
                resUnlock = self.db._query("UNLOCK TABLES", connection)
                return result
            result = self.__rebuildLevelIndexes(parentID, connection)
            resUnlock = self.db._query("UNLOCK TABLES", connection)

        return S_OK()

    def _getConnection(self, connection=False):
        if connection:
            return connection
        res = self.db._getConnection()
        if res["OK"]:
            return res["Value"]
        return connection

    def __rebuildLevelIndexes(self, parentID, connection=False):
        """Rebuild level indexes for all the subdirectories"""
        result = self.__getNumericPath(parentID, connection)
        if not result["OK"]:
            return result

        parentIndexList = result["Value"]
        parentLevel = result["Level"]
        result = self.getChildren(parentID, connection)
        if not result["OK"]:
            return result
        subIDList = result["Value"]

        indexList = list(parentIndexList)
        indexList.append(0)
        for dirID in subIDList:
            indexList[-1] += 1
            lpaths = ["LPATH%d=%d" % (i + 1, indexList[i]) for i in range(parentLevel + 1)]
            lpathString = "SET " + ",".join(lpaths)
            req = "UPDATE FC_DirectoryLevelTree %s WHERE DirID=%s" % (lpathString, dirID)
            result = self.db._update(req, connection)
            if not result["OK"]:
                return result
            result = self.__rebuildLevelIndexes(dirID, connection)

        return S_OK()



""" Test_RSS_Policy_Configurations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id:  $"

import unittest

import DIRAC.ResourceStatusSystem.Policy.Configurations as moduleTested


################################################################################


class Configurations_TestCase(unittest.TestCase):
    def setUp(self):
        """
        Setup
        """

        self.moduleTested = moduleTested

    def tearDown(self):
        """
        TearDown
        """
        del self.moduleTested


################################################################################
# Tests


class Configurations_Success(Configurations_TestCase):
    def test_policiesMeta(self):
        """tests that the configuration does not have any funny key"""

        self.assertEqual(True, hasattr(self.moduleTested, "POLICIESMETA"))

        policiesMeta = self.moduleTested.POLICIESMETA

        for _policyName, policyMeta in policiesMeta.items():
            self.assertEqual({"args", "command", "description", "module"}, set(policyMeta))


################################################################################
################################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" ResourceStatusDB:
    This module provides definition of the DB tables, and methods to access them.

    Written using sqlalchemy declarative_base


    For extending the ResourceStatusDB tables:

    1) In the extended module, call:

    from DIRAC.ResourceStatusSystem.DB.ResourceStatusDB import rmsBase, TABLESLIST, TABLESLISTWITHID
    TABLESLIST = TABLESLIST + [list of new table names]
    TABLESLISTWITHID = TABLESLISTWITHID + [list of new table names]

    2) provide a declarative_base definition of the tables (new or extended) in the extension module

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import six
import datetime
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm.query import Query
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, DateTime, exc, BigInteger

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Core.Base.SQLAlchemyDB import SQLAlchemyDB
from DIRAC.ResourceStatusSystem.Utilities import Utils


TABLESLIST = ["SiteStatus", "ResourceStatus", "NodeStatus"]

TABLESLISTWITHID = [
    "ResourceStatusCache",
    "SiteLog",
    "SiteHistory",
    "ResourceLog",
    "ResourceHistory",
    "NodeLog",
    "NodeHistory",
]


# Defining the tables

rssBase = declarative_base()


class ResourceStatusCache(rssBase):
    """
    Table for EmailAction
    """

    __tablename__ = "ResourceStatusCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    id = Column("ID", BigInteger, nullable=False, autoincrement=True, primary_key=True)
    sitename = Column("SiteName", String(64), nullable=False)
    name = Column("ResourceName", String(64), nullable=False)
    status = Column("Status", String(8), nullable=False, server_default="")
    previousstatus = Column("PreviousStatus", String(8), nullable=False, server_default="")
    statustype = Column("StatusType", String(128), nullable=False, server_default="all")
    time = Column("Time", DateTime, nullable=False, server_default="9999-12-31 23:59:59")

    def fromDict(self, dictionary):
        """
        Fill the fields of the AccountingCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.id = dictionary.get("ID", self.id)
        self.name = dictionary.get("ResourceName", self.name)
        self.sitename = dictionary.get("SiteName", self.sitename)
        self.status = dictionary.get("Status", self.status)
        self.previousstatus = dictionary.get("PreviousStatus", self.previousstatus)
        self.statustype = dictionary.get("StatusType", self.statustype)
        self.time = dictionary.get("Time", datetime.datetime.utcnow())

    def toList(self):
        """
        Simply returns a list of column values
        """

        return [self.id, self.sitename, self.name, self.status, self.previousstatus, self.statustype, self.time]


class ElementStatusBase(object):
    """
    Prototype for tables.
    """

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    name = Column("Name", String(64), nullable=False, primary_key=True)
    statustype = Column("StatusType", String(128), nullable=False, server_default="all", primary_key=True)
    vo = Column("VO", String(64), nullable=False, primary_key=True, server_default="all")
    status = Column("Status", String(8), nullable=False, server_default="")
    reason = Column("Reason", String(512), nullable=False, server_default="Unspecified")
    dateeffective = Column("DateEffective", DateTime, nullable=False)
    tokenexpiration = Column("TokenExpiration", DateTime, nullable=False, server_default="9999-12-31 23:59:59")
    elementtype = Column("ElementType", String(32), nullable=False, server_default="")
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False, server_default="1000-01-01 00:00:00")
    tokenowner = Column("TokenOwner", String(16), nullable=False, server_default="rs_svc")

    columnsOrder = [
        "Name",
        "StatusType",
        "Status",
        "Reason",
        "DateEffective",
        "TokenExpiration",
        "ElementType",
        "LastCheckTime",
        "TokenOwner",
        "VO",
    ]

    def fromDict(self, dictionary):
        """
        Fill the fields of the AccountingCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        utcnow = (
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0)
        )

        self.name = dictionary.get("Name", self.name)
        self.statustype = dictionary.get("StatusType", self.statustype)
        self.vo = dictionary.get("VO", self.vo)
        self.status = dictionary.get("Status", self.status)
        self.reason = dictionary.get("Reason", self.reason)
        self.dateeffective = dictionary.get("DateEffective", self.dateeffective)
        self.tokenexpiration = dictionary.get("TokenExpiration", self.tokenexpiration)
        self.elementtype = dictionary.get("ElementType", self.elementtype)
        self.lastchecktime = dictionary.get("LastCheckTime", utcnow)
        self.tokenowner = dictionary.get("TokenOwner", self.tokenowner)

        if self.dateeffective:
            self.dateeffective = self.dateeffective.replace(microsecond=0)
        if self.tokenexpiration:
            self.tokenexpiration = self.tokenexpiration.replace(microsecond=0)

    def toList(self):
        """Simply returns a list of column values"""
        return [
            self.name,
            self.statustype,
            self.status,
            self.reason,
            self.dateeffective,
            self.tokenexpiration,
            self.elementtype,
            self.lastchecktime,
            self.tokenowner,
            self.vo,
        ]


class ElementStatusBaseWithID(ElementStatusBase):
    """Prototype for tables

    This is almost the same as ElementStatusBase, with the following differences:
    - there's an autoincrement ID column which is also the primary key
    - the name and statusType components are not part of the primary key
    """

    id = Column("ID", BigInteger, nullable=False, autoincrement=True, primary_key=True)
    name = Column("Name", String(64), nullable=False)
    statustype = Column("StatusType", String(128), nullable=False, server_default="all")
    vo = Column("VO", String(64), nullable=False, server_default="all")
    status = Column("Status", String(8), nullable=False, server_default="")
    reason = Column("Reason", String(512), nullable=False, server_default="Unspecified")
    dateeffective = Column("DateEffective", DateTime, nullable=False)
    tokenexpiration = Column("TokenExpiration", DateTime, nullable=False, server_default="9999-12-31 23:59:59")
    elementtype = Column("ElementType", String(32), nullable=False, server_default="")
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False, server_default="1000-01-01 00:00:00")
    tokenowner = Column("TokenOwner", String(16), nullable=False, server_default="rs_svc")

    columnsOrder = [
        "ID",
        "Name",
        "StatusType",
        "Status",
        "Reason",
        "DateEffective",
        "TokenExpiration",
        "ElementType",
        "LastCheckTime",
        "TokenOwner",
        "VO",
    ]

    def fromDict(self, dictionary):
        """
        Fill the fields of the AccountingCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.id = dictionary.get("ID", self.id)
        super(ElementStatusBaseWithID, self).fromDict(dictionary)

    def toList(self):
        """Simply returns a list of column values"""
        return [
            self.id,
            self.name,
            self.statustype,
            self.status,
            self.reason,
            self.dateeffective,
            self.tokenexpiration,
            self.elementtype,
            self.lastchecktime,
            self.tokenowner,
            self.vo,
        ]


# tables with schema defined in ElementStatusBase


class SiteStatus(ElementStatusBase, rssBase):
    """SiteStatus table"""

    __tablename__ = "SiteStatus"


class ResourceStatus(ElementStatusBase, rssBase):
    """ResourceStatusDB table"""

    __tablename__ = "ResourceStatus"


class NodeStatus(ElementStatusBase, rssBase):
    """NodeStatus table"""

    __tablename__ = "NodeStatus"


# tables with schema defined in ElementStatusBaseWithID


class SiteLog(ElementStatusBaseWithID, rssBase):
    """SiteLog table"""

    __tablename__ = "SiteLog"


class SiteHistory(ElementStatusBaseWithID, rssBase):
    """SiteHistory table"""

    __tablename__ = "SiteHistory"


class ResourceLog(ElementStatusBaseWithID, rssBase):
    """ResourceLog table"""

    __tablename__ = "ResourceLog"


class ResourceHistory(ElementStatusBaseWithID, rssBase):
    """ResourceHistory table"""

    __tablename__ = "ResourceHistory"


class NodeLog(ElementStatusBaseWithID, rssBase):
    """NodeLog table"""

    __tablename__ = "NodeLog"


class NodeHistory(ElementStatusBaseWithID, rssBase):
    """NodeHistory table"""

    __tablename__ = "NodeHistory"


# Interaction with the DB


class ResourceStatusDB(SQLAlchemyDB):
    """
    Class that defines the interactions with the tables of the ResourceStatusDB.
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        super(ResourceStatusDB, self).__init__()

        # These are the list of tables that will be created.
        # They can be extended in an extension module
        self.tablesList = getattr(Utils.voimport("DIRAC.ResourceStatusSystem.DB.ResourceStatusDB"), "TABLESLIST")
        self.tablesListWithID = getattr(
            Utils.voimport("DIRAC.ResourceStatusSystem.DB.ResourceStatusDB"), "TABLESLISTWITHID"
        )

        self.extensions = gConfig.getValue("DIRAC/Extensions", [])
        self._initializeConnection("ResourceStatus/ResourceStatusDB")

        # Create required tables
        self._createTablesIfNotThere(self.tablesList)
        self._createTablesIfNotThere(self.tablesListWithID)

    def insert(self, table, params):
        """
        Inserts params in the DB.

        :param table: table where to insert
        :type table: str
        :param params: Dictionary to fill a single line
        :type params: dict

        :return: S_OK() || S_ERROR()
        """

        if not params.get("DateEffective"):
            params["DateEffective"] = datetime.datetime.utcnow().replace(microsecond=0)

        return super(ResourceStatusDB, self).insert(table, params)

    def addOrModify(self, table, params):
        """
        Using the PrimaryKeys of the table, it looks for the record in the database.
        If it is there, it is updated, if not, it is inserted as a new entry.

        :param table: table where to add or modify
        :type table: str
        :param params: dictionary of what to add or modify
        :type params: dict

        :return: S_OK() || S_ERROR()
        """

        session = self.sessionMaker_o()
        found = False
        for ext in self.extensions:
            try:
                table_c = getattr(__import__(ext + __name__, globals(), locals(), [table]), table)
                found = True
                break
            except (ImportError, AttributeError):
                continue
        # If not found in extensions, import it from DIRAC base (this same module).
        if not found:
            table_c = getattr(__import__(__name__, globals(), locals(), [table]), table)
        primaryKeys = [key.name for key in class_mapper(table_c).primary_key]

        try:
            select = Query(table_c, session=session)
            for columnName, columnValue in params.items():
                if not columnValue or columnName not in primaryKeys:
                    continue
                column_a = getattr(table_c, columnName.lower())
                if isinstance(columnValue, (list, tuple)):
                    select = select.filter(column_a.in_(list(columnValue)))
                elif isinstance(columnValue, six.string_types):
                    select = select.filter(column_a == columnValue)
                else:
                    self.log.error("type(columnValue) == %s" % type(columnValue))

            res = select.first()  # the selection is done via primaryKeys only
            if not res:  # if not there, let's insert it (and exit)
                return self.insert(table, params)

            # From now on, we assume we need to modify

            # Treating case of time value updates
            if not params.get("LastCheckTime"):
                params["LastCheckTime"] = None
            if not params.get("DateEffective"):
                params["DateEffective"] = None

            # Should we change DateEffective?
            changeDE = False
            if params.get("Status"):
                if params.get("Status") != res.status:  # we update dateEffective iff we change the status
                    changeDE = True

            for columnName, columnValue in params.items():
                if columnName == "LastCheckTime" and not columnValue:  # we always update lastCheckTime
                    columnValue = datetime.datetime.utcnow().replace(microsecond=0)
                if changeDE and columnName == "DateEffective" and not columnValue:
                    columnValue = datetime.datetime.utcnow().replace(microsecond=0)
                if columnValue:
                    if isinstance(columnValue, datetime.datetime):
                        columnValue = columnValue.replace(microsecond=0)
                    setattr(res, columnName.lower(), columnValue)
            session.commit()

            # and since we modified, we now insert a new line in the log table
            return self.insert(table.replace("Status", "") + "Log", params)
            # The line inserted will maybe become a History line thanks to the SummarizeLogsAgent

        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("addOrModify: unexpected exception", lException=e)
            return S_ERROR("addOrModify: unexpected exception %s" % e)
        finally:
            session.close()

    def addIfNotThere(self, table, params):
        """
        Using the PrimaryKeys of the table, it looks for the record in the database.
        If it is not there, it is inserted as a new entry.

        :param table: table where to add or modify
        :type table: str
        :param params: dictionary of what to add or modify
        :type params: dict

        :return: S_OK() || S_ERROR()
        """

        session = self.sessionMaker_o()
        table_c = getattr(__import__(__name__, globals(), locals(), [table]), table)
        primaryKeys = [key.name for key in class_mapper(table_c).primary_key]

        try:
            select = Query(table_c, session=session)
            for columnName, columnValue in params.items():
                if not columnValue or columnName not in primaryKeys:
                    continue
                column_a = getattr(table_c, columnName.lower())
                if isinstance(columnValue, (list, tuple)):
                    select = select.filter(column_a.in_(list(columnValue)))
                elif isinstance(columnValue, six.string_types):
                    select = select.filter(column_a == columnValue)
                else:
                    self.log.error("type(columnValue) == %s" % type(columnValue))

            res = select.first()  # the selection is done via primaryKeys only
            if not res:  # if not there, let's insert it
                return self.insert(table, params)

            session.commit()
            return S_OK()

        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("addIfNotThere: unexpected exception", lException=e)
            return S_ERROR("addIfNotThere: unexpected exception %s" % e)
        finally:
            session.close()



# ##WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING #
#                                           Under development                                                   #
# ##WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING###WARNING #

""" Module to upload specified job output files according to the parameters
    defined in the production workflow.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import gLogger
from DIRAC.Workflow.Modules.ModuleBase import ModuleBase, GracefulTermination


class UploadOutputs(ModuleBase):

    #############################################################################

    def __init__(self):
        """c'tor"""
        self.log = gLogger.getSubLogger("UploadOutputs")
        super(UploadOutputs, self).__init__(self.log)

        self.outputDataStep = ""
        self.outputData = None
        self.outputList = []
        self.defaultOutputSE = []
        self.outputSE = []
        self.outputPath = ""

    #############################################################################

    def _resolveInputVariables(self):
        """The module parameters are resolved here."""
        super(UploadOutputs, self)._resolveInputVariables()

        # this comes from Job().setOutputData(). Typical for user jobs
        if "OutputData" in self.workflow_commons:
            self.outputData = self.workflow_commons["OutputData"]
            if isinstance(self.outputData, six.string_types):
                self.outputData = [i.strip() for i in self.outputData.split(";")]
        # if not present, we use the outputList, which is instead incrementally created based on the single step outputs
        # This is more typical for production jobs, that can have many steps linked one after the other
        elif "outputList" in self.workflow_commons:
            self.outputList = self.workflow_commons["outputList"]
        else:
            raise GracefulTermination("Nothing to upload")

        # in case you want to put a mask on the steps
        # TODO: add it to the DIRAC API
        if "outputDataStep" in self.workflow_commons:
            self.outputDataStep = self.workflow_commons["outputDataStep"]

        # this comes from Job().setOutputData(). Typical for user jobs
        if "OutputSE" in self.workflow_commons:
            specifiedSE = self.workflow_commons["OutputSE"]
            if not isinstance(specifiedSE, list):
                self.outputSE = [i.strip() for i in specifiedSE.split(";")]
        else:
            self.log.verbose("No OutputSE specified, using default value: %s" % (", ".join(self.defaultOutputSE)))

        # this comes from Job().setOutputData(). Typical for user jobs
        if "OutputPath" in self.workflow_commons:
            self.outputPath = self.workflow_commons["OutputPath"]

    def _initialize(self):
        """gets the files to upload, check if to upload"""
        # lfnsList = self.__getOutputLFNs( self.outputData ) or outputList?

        if not self._checkWFAndStepStatus():
            raise GracefulTermination("No output data upload attempted")

    def __getOuputLFNs(self, outputList, *args):
        """This is really VO-specific.
        It should be replaced by each VO. Setting an LFN here just as an idea, and for testing purposes.
        """
        lfnList = []
        for outputFile in outputList:
            lfnList.append("/".join([str(x) for x in args]) + outputFile)

        return lfnList

    def _execute(self):
        """uploads the files"""
        pass



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-proxy-upload.py
# Author :  Adrian Casajus
########################################################################
"""
Upload proxy.

Example:
  $ dirac-admin-proxy-upload
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.ProxyUpload import CLIParams, uploadProxy


@Script()
def main():
    cliParams = CLIParams()
    cliParams.registerCLISwitches()

    Script.parseCommandLine()

    retVal = uploadProxy(cliParams)
    if not retVal["OK"]:
        print(retVal["Message"])
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :   tornado-start-all
# Author : Louis MARTIN
########################################################################
# Just run this script to start Tornado and all services
# Use CS to change port

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():

    if os.environ.get("DIRAC_USE_TORNADO_IOLOOP", "false").lower() not in ("yes", "true"):
        raise RuntimeError(
            "DIRAC_USE_TORNADO_IOLOOP is not defined in the environment."
            + "\n"
            + "It is necessary to run with Tornado."
            + "\n"
            + "https://dirac.readthedocs.io/en/latest/DeveloperGuide/TornadoServices/index.html"
        )

    from DIRAC import gConfig
    from DIRAC.ConfigurationSystem.Client import PathFinder
    from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
    from DIRAC.Core.Tornado.Server.TornadoServer import TornadoServer
    from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
    from DIRAC.FrameworkSystem.Client.Logger import gLogger

    localCfg = Script.localCfg
    localCfg.setConfigurationForServer("Tornado/Tornado")
    localCfg.addMandatoryEntry("/DIRAC/Setup")
    localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
    localCfg.addDefaultEntry("LogLevel", "INFO")
    localCfg.addDefaultEntry("LogColor", True)
    resultDict = localCfg.loadUserData()
    if not resultDict["OK"]:
        gLogger.initialize("Tornado", "/")
        gLogger.error("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    includeExtensionErrors()

    gLogger.initialize("Tornado", "/")

    # We check if there is no configuration server started as master
    # If you want to start a master CS you should use Configuration_Server.cfg and
    # use tornado-start-CS.py
    key = "/Systems/Configuration/%s/Services/Server/Protocol" % PathFinder.getSystemInstance("Configuration")
    if gConfigurationData.isMaster() and gConfig.getValue(key, "dips").lower() == "https":
        gLogger.fatal("You can't run the CS and services in the same server!")
        sys.exit(0)

    serverToLaunch = TornadoServer()
    serverToLaunch.startTornado()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author: lintao

"""
This module is a helper module for command line interface.

It can help you auto complete the directory name.

The code is a part of

  https://github.com/mirguest/CmdDirectoryCompletion

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-sync-users-from-file
# Author :  Adrian Casajus
########################################################################
"""
Sync users in Configuration with the cfg contents.

Example:
  $ dirac-admin-sync-users-from-file file_users.cfg
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from diraccfg import CFG

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"


@Script()
def main():
    Script.registerSwitch("t", "test", "Only test. Don't commit changes")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        "UserCfg:  Cfg FileName with Users as sections containing" "DN, Groups, and other properties as options"
    )
    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getExtraCLICFGFiles()

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    testOnly = False
    errorList = []

    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("t", "test"):
            testOnly = True

    try:
        usersCFG = CFG().loadFromFile(args[0])
    except Exception as e:
        errorList.append("file open", "Can't parse file %s: %s" % (args[0], str(e)))
        errorCode = 1
    else:
        if not diracAdmin.csSyncUsersWithCFG(usersCFG):
            errorList.append(("modify users", "Cannot sync with %s" % args[0]))
            exitCode = 255

    if not exitCode and not testOnly:
        result = diracAdmin.csCommitChanges()
        if not result["OK"]:
            errorList.append(("commit", result["Message"]))
            exitCode = 255

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#########################################################################
#  Host.py
#  4.11.2014
#  Author: A.T.
#########################################################################

""" Host - class for managing jobs on a host. Host objects are invoked
    with LocalComputingElement or SSHComputingElement objects
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import glob
import shutil
import signal

try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess
import stat
import json
import multiprocessing
from datetime import datetime, timedelta

__RCSID__ = "$Id$"

# Clean job info and output after so many days
CLEAN_DELAY = timedelta(7)


class Host(object):
    def __init__(self):
        self.nCores = 1
        try:
            self.nCores = multiprocessing.cpu_count()
        except Exception:
            pass

    def submitJob(self, **kwargs):

        resultDict = {}

        args = dict(kwargs)

        MANDATORY_PARAMETERS = [
            "Executable",
            "SharedDir",
            "OutputDir",
            "ErrorDir",
            "WorkDir",
            "InfoDir",
            "ExecutionContext",
            "JobStamps",
        ]

        for argument in MANDATORY_PARAMETERS:
            if argument not in args:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = args.get("NJobs", 1)
        stamps = args["JobStamps"]
        context = args.get("ExecutionContext", "Local")
        jobidName = context.upper() + "_JOBID"
        nCores = args.get("NCores", 1)
        nodeHost = args.get("SSHNodeHost")

        # Prepare the executor command
        runFileName = os.path.join(args["SharedDir"], "run_detached.sh")
        runFileName = os.path.expandvars(runFileName)
        if os.path.isfile(runFileName):
            os.unlink(runFileName)
        runFile = open(runFileName, "w")
        runFile.write(
            """
( exec </dev/null
#  echo $2
  exec > $2
#  echo $3
  exec 2> $3
#  echo $1
  exec setsid $1
) &
kill -0 $! > /dev/null 2>&1 || exit 1
echo $!
exit 0
"""
        )
        runFile.close()
        os.chmod(runFileName, stat.S_IXUSR | stat.S_IRUSR)
        jobs = []
        output = ""
        args["RunFile"] = runFileName
        for _i in range(int(nJobs)):
            args["Stamp"] = stamps[_i]
            envDict = os.environ
            envDict[jobidName] = stamps[_i]
            if nodeHost:
                envDict["SSH_NODE_HOST"] = nodeHost
            try:
                jobDir = "%(WorkDir)s/%(Stamp)s" % args
                jobDir = os.path.expandvars(jobDir)
                os.makedirs(jobDir)
                os.chdir(jobDir)
                popenObject = subprocess.Popen(
                    ["%(RunFile)s %(Executable)s %(OutputDir)s/%(Stamp)s.out %(ErrorDir)s/%(Stamp)s.err" % args],
                    stdout=subprocess.PIPE,
                    shell=True,
                    env=envDict,
                    universal_newlines=True,
                )
                pid = popenObject.communicate()[0]
            except OSError as x:
                output = str(x)
                break
            pid = int(pid)
            if pid:
                # Store the job info
                jobInfo = {
                    "PID": pid,
                    "SubmissionTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                    "JOBID": stamps[_i],
                    "NCores": nCores,
                }
                jobString = json.dumps(jobInfo)
                pidFileName = "%(InfoDir)s/%(Stamp)s.info" % args
                pidFileName = os.path.expandvars(pidFileName)
                with open(pidFileName, "w") as pd:
                    pd.write(jobString)
                jobs.append(stamps[_i])
            else:
                break

        if jobs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = jobs
        else:
            resultDict["Status"] = 1
            resultDict["Message"] = output

        return resultDict

    def __cleanJob(self, stamp, infoDir, workDir, outputDir=None, errorDir=None):

        jobDir = os.path.join(workDir, stamp)
        if os.path.isdir(jobDir):
            shutil.rmtree(jobDir)
        pidFile = os.path.join(infoDir, "%s.info" % stamp)
        if os.path.isfile(pidFile):
            os.unlink(pidFile)
        if outputDir:
            outFile = os.path.join(outputDir, "%s.out" % stamp)
            if os.path.isfile(outFile):
                os.unlink(outFile)
        if errorDir:
            errFile = os.path.join(errorDir, "%s.err" % stamp)
            if os.path.isfile(errFile):
                os.unlink(errFile)

    def __getJobInfo(self, infoDir, stamp):

        jobInfo = {}
        infoFileName = os.path.join(infoDir, "%s.info" % stamp)
        infoFileName = os.path.expandvars(infoFileName)
        if os.path.exists(infoFileName):
            infoFile = open(infoFileName, "r")
            jobInfo = infoFile.read().strip()
            infoFile.close()
            jobInfo = json.loads(jobInfo)
        return jobInfo

    def getCEStatus(self, **kwargs):
        """Get the overall CE status"""
        resultDict = {"Running": 0, "Waiting": 0}

        MANDATORY_PARAMETERS = ["InfoDir", "WorkDir", "OutputDir", "ErrorDir", "User"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        user = kwargs.get("User")
        infoDir = kwargs.get("InfoDir")
        workDir = kwargs.get("WorkDir")
        outputDir = kwargs.get("OutputDir")
        errorDir = kwargs.get("ErrorDir")

        running = 0
        usedCores = 0
        infoDir = os.path.expandvars(infoDir)
        infoFiles = glob.glob("%s/*.info" % infoDir)
        for infoFileName in infoFiles:
            infoFileName = os.path.expandvars(infoFileName)
            infoFile = open(infoFileName, "r")
            jobInfo = infoFile.read().strip()
            infoFile.close()
            jobInfo = json.loads(jobInfo)
            pid = jobInfo["PID"]
            sp = subprocess.Popen(
                ["ps", "-f", "-p", str(pid), "--no-headers"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            if len(output.split("\n")) == 2 and "wrapper" in output:
                running += 1
                usedCores += jobInfo["NCores"]
            else:
                stamp = jobInfo["JOBID"]
                jobLife = datetime.utcnow() - datetime.strptime(jobInfo["SubmissionTime"], "%Y-%m-%d %H:%M:%S")
                if jobLife > CLEAN_DELAY:
                    self.__cleanJob(stamp, infoDir, workDir, outputDir, errorDir)

        resultDict["Status"] = 0
        resultDict["Running"] = running
        availableCores = self.nCores - usedCores
        resultDict["AvailableCores"] = availableCores
        return resultDict

    def __checkPid(self, pid, user):

        if pid == 0:
            return "Unknown"
        sp = subprocess.Popen(
            ["ps", "-f", "-p", str(pid), "--no-headers"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode
        if status == 0 and len(output.split("\n")) == 2 and user in output:
            return "Running"
        return "Done"

    def getJobStatus(self, **kwargs):

        resultDict = {}

        MANDATORY_PARAMETERS = ["InfoDir", "JobIDList", "User"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        user = kwargs.get("User")
        infoDir = kwargs.get("InfoDir")
        jobStamps = kwargs.get("JobIDList")
        jobDict = {}
        for stamp in jobStamps:
            pid = self.__getJobInfo(infoDir, stamp).get("PID", 0)
            jobDict[stamp] = self.__checkPid(pid, user)

        resultDict["Status"] = 0
        resultDict["Jobs"] = jobDict
        return resultDict

    def killJob(self, **kwargs):

        resultDict = {}

        MANDATORY_PARAMETERS = ["InfoDir", "WorkDir", "OutputDir", "ErrorDir", "JobIDList", "User"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        user = kwargs.get("User")
        infoDir = kwargs.get("InfoDir")
        workDir = kwargs.get("WorkDir")
        outputDir = kwargs.get("OutputDir")
        errorDir = kwargs.get("ErrorDir")
        jobStamps = kwargs.get("JobIDList")
        jobDict = {}
        for stamp in jobStamps:
            pid = self.__getJobInfo(infoDir, stamp).get("PID", 0)
            if self.__checkPid(pid, user) == "Running":
                os.kill(pid, signal.SIGKILL)
                self.__cleanJob(stamp, infoDir, workDir, outputDir, errorDir)
                jobDict[stamp] = "Killed"
            else:
                jobDict[stamp] = "Done"

        resultDict["Status"] = 0
        resultDict["Successful"] = jobStamps
        resultDict["Failed"] = []
        resultDict["Jobs"] = jobDict
        return resultDict



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import base64
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DEncode

gForceRawEncoding = False

try:
    import zlib

    gZCompressionEnabled = True
except ImportError as x:
    gZCompressionEnabled = False


def codeRequestInFileId(plotRequest, compressIfPossible=True):
    compress = compressIfPossible and gZCompressionEnabled
    thbStub = False
    if compress:
        plotStub = "Z:%s" % base64.urlsafe_b64encode(zlib.compress(DEncode.encode(plotRequest), 9)).decode()
    elif not gForceRawEncoding:
        plotStub = "S:%s" % base64.urlsafe_b64encode(DEncode.encode(plotRequest))
    else:
        plotStub = "R:%s" % DEncode.encode(plotRequest)
    # If thumbnail requested, use plot as thumbnail, and generate stub for plot without one
    extraArgs = plotRequest["extraArgs"]
    if "thumbnail" in extraArgs and extraArgs["thumbnail"]:
        thbStub = plotStub
        extraArgs["thumbnail"] = False
        if compress:
            plotStub = "Z:%s" % base64.urlsafe_b64encode(zlib.compress(DEncode.encode(plotRequest), 9)).decode()
        elif not gForceRawEncoding:
            plotStub = "S:%s" % base64.urlsafe_b64encode(DEncode.encode(plotRequest)).decode()
        else:
            plotStub = "R:%s" % DEncode.encode(plotRequest).decode()
    return S_OK({"plot": plotStub, "thumbnail": thbStub})


def extractRequestFromFileId(fileId):
    stub = fileId[2:]
    compressType = fileId[0]
    if compressType == "Z":
        gLogger.info("Compressed request, uncompressing")
        try:
            # Encoding is only required for Python 2 and can be removed when Python 2 support is no longer needed
            stub = base64.urlsafe_b64decode(stub.encode())
        except Exception as e:
            gLogger.error("Oops! Plot request is not properly encoded!", str(e))
            return S_ERROR("Oops! Plot request is not properly encoded!: %s" % str(e))
        try:
            stub = zlib.decompress(stub)
        except Exception as e:
            gLogger.error("Oops! Plot request is invalid!", str(e))
            return S_ERROR("Oops! Plot request is invalid!: %s" % str(e))
    elif compressType == "S":
        gLogger.info("Base64 request, decoding")
        try:
            stub = base64.urlsafe_b64decode(stub)
        except Exception as e:
            gLogger.error("Oops! Plot request is not properly encoded!", str(e))
            return S_ERROR("Oops! Plot request is not properly encoded!: %s" % str(e))
    elif compressType == "R":
        # Do nothing, it's already uncompressed
        pass
    else:
        gLogger.error("Oops! Stub type is unknown", compressType)
        return S_ERROR("Oops! Stub type '%s' is unknown :P" % compressType)
    plotRequest, stubLength = DEncode.decode(stub)
    if len(stub) != stubLength:
        gLogger.error("Oops! The stub is longer than the data :P")
        return S_ERROR("Oops! The stub is longer than the data :P")
    return S_OK(plotRequest)



"""
Defines the plugin to take storage space information given by WLCG Accounting Json
https://twiki.cern.ch/twiki/bin/view/LCG/AccountingTaskForce#Storage_Space_Accounting
https://twiki.cern.ch/twiki/pub/LCG/AccountingTaskForce/storage_service_v4.txt
https://docs.google.com/document/d/1yzCvKpxsbcQC5K9MyvXc-vBF1HGPBk4vhjw3MEXoXf8

When this is used, the OccupancyLFN has to be the full HTTP(s) URL

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import requests

from DIRAC.Core.Security.Locations import getCAsLocation
from DIRAC.Resources.Storage.OccupancyPlugins.WLCGAccountingJson import WLCGAccountingJson


class WLCGAccountingHTTPJson(WLCGAccountingJson):
    """.. class:: WLCGAccountingHTTPJson

    Occupancy plugin to return the space information given by WLCG HTTP Accounting Json

    """

    def __init__(self, se):
        """
        c'tor

        :param se: reference to the StorageElement object from which we are called
        """

        super(WLCGAccountingHTTPJson, self).__init__(se)

        self.log = se.log.getSubLogger("WLCGAccountingHTTPJson")

    def _downloadJsonFile(self, occupancyLFN, filePath):
        """Download the json file at the location using requests

        :param occupancyLFN: this is actually a full https URL
        :param filePath: destination path for the file

        """

        try:
            with open(filePath, "wt") as fd:
                caPath = getCAsLocation()
                res = requests.get(occupancyLFN, verify=caPath)
                res.raise_for_status()
                fd.write(res.text)
        except Exception as e:
            self.log.debug("Exception while copying", repr(e))
            os.remove(filePath)



"""Unit tests of MQProducer interface in the DIRAC.Resources.MessageQueue.MProducerQ
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from DIRAC import S_OK
from DIRAC.Resources.MessageQueue.MQProducer import MQProducer
from DIRAC.Resources.MessageQueue.MQConnectionManager import MQConnectionManager
from DIRAC.Resources.MessageQueue.MQConnector import MQConnector


class FakeMQConnector(MQConnector):
    def __init__(self, params={}):
        super(FakeMQConnector, self).__init__()

    def disconnect(self):
        return S_OK("FakeMQConnection disconnecting")

    def get(self, destination=""):
        return "FakeMQConnection getting message"

    def put(self, message, parameters=None):
        return S_OK("FakeMQConnection sending message: " + str(message))


class TestMQProducer(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None  # To show full difference between structures in  case of error
        dest = {}
        dest.update({"/queue/FakeQueue": ["producer4", "producer2"]})
        dest4 = {"/queue/test3": ["producer1", "consumer2", "consumer3", "consumer4"]}
        conn1 = {"MQConnector": FakeMQConnector(), "destinations": dest}
        conn2 = {"MQConnector": FakeMQConnector(), "destinations": dest4}
        storage = {"fake.cern.ch": conn1, "testdir.blabla.ch": conn2}
        self.myManager = MQConnectionManager(connectionStorage=storage)

    def tearDown(self):
        pass


class TestMQProducer_put(TestMQProducer):
    def test_success(self):
        producer = MQProducer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", producerId="producer4")
        result = producer.put("wow!")
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "FakeMQConnection sending message: wow!")

    def test_failure(self):
        producer = MQProducer(mqManager=self.myManager, mqURI="bad.cern.ch::Queues::FakeQueue", producerId="producer4")
        result = producer.put("wow!")
        self.assertFalse(result["OK"])


class TestMQProducer_close(TestMQProducer):
    def test_success(self):
        producer = MQProducer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", producerId="producer4")
        result = producer.close()
        self.assertTrue(result["OK"])
        # producer is still able to sent cause the connection is still active (producer2 is connected)
        result = producer.put("wow!")
        self.assertTrue(result["OK"])

    def test_failure(self):
        producer = MQProducer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", producerId="producer4")
        result = producer.close()
        self.assertTrue(result["OK"])
        result = producer.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger producer4 does not exist!)",
        )

    def test_failure2(self):
        producer = MQProducer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", producerId="producer4")
        producer2 = MQProducer(
            mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", producerId="producer2"
        )
        result = producer.close()
        self.assertTrue(result["OK"])
        result = producer.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger producer4 does not exist!)",
        )
        result = producer2.close()
        self.assertTrue(result["OK"])
        result = producer2.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger producer2 does not exist!)",
        )
        # connection does not exist so put will not work
        result = producer.put("wow!")
        self.assertFalse(result["OK"])
        self.assertEqual(result["Message"], "Failed to get the MQConnector!")


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestMQProducer)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQProducer_put))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQProducer_close))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-proxy
# Author :  Stuart Paterson
########################################################################
"""
Retrieve a delegated proxy for the given user and group

Example:
  $ dirac-admin-get-proxy vhamar dirac_user
  Proxy downloaded to /afs/in2p3.fr/home/h/hamar/proxy.vhamar.dirac_user
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

import six

import DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.Helpers import Registry

__RCSID__ = "$Id$"


class Params(object):

    limited = False
    proxyPath = False
    proxyLifeTime = 86400
    enableVOMS = False
    vomsAttr = None

    def setLimited(self, args):
        """Set limited

        :param bool args: is limited

        :return: S_OK()/S_ERROR()
        """
        self.limited = True
        return S_OK()

    def setProxyLocation(self, args):
        """Set proxy location

        :param str args: proxy path

        :return: S_OK()/S_ERROR()
        """
        self.proxyPath = args
        return S_OK()

    def setProxyLifeTime(self, arg):
        """Set proxy lifetime

        :param int arg: lifetime in a seconds

        :return: S_OK()/S_ERROR()
        """
        try:
            fields = [f.strip() for f in arg.split(":")]
            self.proxyLifeTime = int(fields[0]) * 3600 + int(fields[1]) * 60
        except Exception:
            gLogger.notice("Can't parse %s time! Is it a HH:MM?" % arg)
            return S_ERROR("Can't parse time argument")
        return S_OK()

    def automaticVOMS(self, arg):
        """Enable VOMS

        :param bool arg: enable VOMS

        :return: S_OK()/S_ERROR()
        """
        self.enableVOMS = True
        return S_OK()

    def setVOMSAttr(self, arg):
        """Register CLI switches

        :param str arg: VOMS attribute
        """
        self.enableVOMS = True
        self.vomsAttr = arg
        return S_OK()

    def registerCLISwitches(self):
        """Register CLI switches"""
        Script.registerSwitch(
            "v:", "valid=", "Valid HH:MM for the proxy. By default is 24 hours", self.setProxyLifeTime
        )
        Script.registerSwitch("l", "limited", "Get a limited proxy", self.setLimited)
        Script.registerSwitch("u:", "out=", "File to write as proxy", self.setProxyLocation)
        Script.registerSwitch(
            "a", "voms", "Get proxy with VOMS extension mapped to the DIRAC group", self.automaticVOMS
        )
        Script.registerSwitch("m:", "vomsAttr=", "VOMS attribute to require", self.setVOMSAttr)


@Script()
def main():
    params = Params()
    params.registerCLISwitches()
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        ("DN:       DN of the user", "user:     DIRAC user name (will fail if there is more than 1 DN registered)")
    )
    Script.registerArgument(" group:    DIRAC group name")

    Script.parseCommandLine(ignoreErrors=True)
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    userDN, userGroup = Script.getPositionalArgs(group=True)

    userName = False
    if userDN.find("/") != 0:
        userName = userDN
        retVal = Registry.getDNForUsername(userName)
        if not retVal["OK"]:
            gLogger.notice("Cannot discover DN for username %s\n\t%s" % (userName, retVal["Message"]))
            DIRAC.exit(2)
        DNList = retVal["Value"]
        if len(DNList) > 1:
            gLogger.notice("Username %s has more than one DN registered" % userName)
            ind = 0
            for dn in DNList:
                gLogger.notice("%d %s" % (ind, dn))
                ind += 1
            inp = six.moves.input("Which DN do you want to download? [default 0] ")
            if not inp:
                inp = 0
            else:
                inp = int(inp)
            userDN = DNList[inp]
        else:
            userDN = DNList[0]

    if not params.proxyPath:
        if not userName:
            result = Registry.getUsernameForDN(userDN)
            if not result["OK"]:
                gLogger.notice("DN '%s' is not registered in DIRAC" % userDN)
                DIRAC.exit(2)
            userName = result["Value"]
        params.proxyPath = "%s/proxy.%s.%s" % (os.getcwd(), userName, userGroup)

    if params.enableVOMS:
        result = gProxyManager.downloadVOMSProxy(
            userDN,
            userGroup,
            limited=params.limited,
            requiredTimeLeft=params.proxyLifeTime,
            requiredVOMSAttribute=params.vomsAttr,
        )
    else:
        result = gProxyManager.downloadProxy(
            userDN, userGroup, limited=params.limited, requiredTimeLeft=params.proxyLifeTime
        )
    if not result["OK"]:
        gLogger.notice("Proxy file cannot be retrieved: %s" % result["Message"])
        DIRAC.exit(2)
    chain = result["Value"]
    result = chain.dumpAllToFile(params.proxyPath)
    if not result["OK"]:
        gLogger.notice("Proxy file cannot be written to %s: %s" % (params.proxyPath, result["Message"]))
        DIRAC.exit(2)
    gLogger.notice("Proxy downloaded to %s" % params.proxyPath)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC.Core.Utilities.Decorators import executeOnlyIf


class AnyClass(object):
    def __init__(self):
        self.attr = None

    @executeOnlyIf("attr", False)
    def withoutSpecificValue(self):
        """Decorator with no specific value"""
        return True

    @executeOnlyIf("undefined", False)
    def withoutSpecificValueUndefined(self):
        """Decorator with no specific value on undefined attribute"""
        return True

    @executeOnlyIf("attr", False, attrVal="specific")
    def withSpecificValue(self):
        """Decorator with specific value"""
        return True

    @executeOnlyIf("undefined", False, attrVal="specific")
    def withSpecificValueUndefined(self):
        """Decorator with specific value on undefined attribute"""
        return True

    @executeOnlyIf("attr1", "attr1Missing")
    @executeOnlyIf("attr2", "attr2Missing")
    def doubleDeco(self):
        """Double deco"""
        return True


# Undefined attribute


def test_without_specific_value_undefined():
    """Test on undefined value"""
    a = AnyClass()
    assert a.withoutSpecificValueUndefined() is False


def test_with_specific_value_undefined():
    """Test on undefined value"""
    a = AnyClass()
    assert a.withSpecificValueUndefined() is False


# Not initialized


def test_without_specific_value_not_initialized():
    """Test when the value was not initialized"""
    a = AnyClass()
    assert a.withoutSpecificValue() is False


def test_with_specific_value_not_initialized():
    """Test when the value was not initialized"""
    a = AnyClass()
    assert a.withSpecificValue() is False


def test_without_specific_value_initialized():
    """Test when the value was initialized to anything that eval to True"""
    a = AnyClass()

    a.attr = True
    assert a.withoutSpecificValue()

    a.attr = 1
    assert a.withoutSpecificValue()

    a.attr = "str"
    assert a.withoutSpecificValue()


def test_with_specific_value_initialized():
    """Test when the value was initialized to anything that eval to True but not what we want"""
    a = AnyClass()

    a.attr = True
    assert a.withSpecificValue() is False

    a.attr = 1
    assert a.withSpecificValue() is False

    a.attr = "str"
    assert a.withSpecificValue() is False


def test_without_specific_value_bad_initialized():
    """Test when the value was initialized but with something that evaluate to False"""
    a = AnyClass()

    a.attr = False
    assert a.withoutSpecificValue() is False

    a.attr = []
    assert a.withoutSpecificValue() is False


def test_with_specific_value_correct():
    """Test when the value was not initialized to anything that should evaluate to True"""
    a = AnyClass()
    a.attr = "specific"
    assert a.withSpecificValue()


def test_doubleDeco():
    """Test when the value was not initialized to anything that should evaluate to True"""
    a = AnyClass()

    assert a.doubleDeco() == "attr1Missing"

    a.attr2 = True
    assert a.doubleDeco() == "attr1Missing"

    a.attr2 = False
    a.attr1 = True

    assert a.doubleDeco() == "attr2Missing"

    a.attr2 = True
    assert a.doubleDeco()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.JEncode import encode, decode


@createClient("DataManagement/FTS3Manager")
class FTS3Client(Client):
    """Client code to the FTS3 service"""

    def __init__(self, url=None, **kwargs):
        """Constructor function."""
        super(FTS3Client, self).__init__(**kwargs)
        self.setServer("DataManagement/FTS3Manager")
        if url:
            self.setServer(url)

    def persistOperation(self, opObj, **kwargs):
        """Persist (insert/update) an FTS3Operation object into the db

        :param opObj: instance of FTS3Operation
        """

        # In case someone manually set sourceSEs as a list:
        if isinstance(opObj.sourceSEs, list):
            opObj.sourceSEs = ",".join(opObj.sourceSEs)

        opJSON = encode(opObj)
        return self._getRPC(**kwargs).persistOperation(opJSON)

    def getOperation(self, operationID, **kwargs):
        """Get the FTS3Operation from the database

        :param operationID: id of the operation
        :return: FTS3Operation object
        """
        res = self._getRPC(**kwargs).getOperation(operationID)
        if not res["OK"]:
            return res

        opJSON = res["Value"]

        try:
            opObj, _size = decode(opJSON)
            return S_OK(opObj)
        except Exception as e:
            return S_ERROR("Exception when decoding the FTS3Operation object %s" % e)

    def getActiveJobs(self, limit=20, lastMonitor=None, jobAssignmentTag="Assigned", **kwargs):
        """Get all the FTSJobs that are not in a final state

        :param limit: max number of jobs to retrieve
        :return: list of FTS3Jobs
        """
        res = self._getRPC(**kwargs).getActiveJobs(limit, lastMonitor, jobAssignmentTag)
        if not res["OK"]:
            return res

        activeJobsJSON = res["Value"]

        try:
            activeJobs, _size = decode(activeJobsJSON)
            return S_OK(activeJobs)
        except Exception as e:
            return S_ERROR("Exception when decoding the active jobs json %s" % e)

    def getNonFinishedOperations(self, limit=20, operationAssignmentTag="Assigned", **kwargs):
        """Get all the FTS3Operations that have files in New or Failed state
        (reminder: Failed is NOT terminal for files. Failed is when fts failed, but we
        can retry)

        :param limit: max number of jobs to retrieve
        :return: json list of FTS3Operation
        """

        res = self._getRPC(**kwargs).getNonFinishedOperations(limit, operationAssignmentTag)
        if not res["OK"]:
            return res

        operationsJSON = res["Value"]

        try:
            operations, _size = decode(operationsJSON)
            return S_OK(operations)
        except Exception as e:
            return S_ERROR(0, "Exception when decoding the non finished operations json %s" % e)

    def getOperationsFromRMSOpID(self, rmsOpID, **kwargs):
        """Get the FTS3Operations matching a given RMS Operation

        :param rmsOpID: id of the operation in the RMS
        :return: list of FTS3Operation objects
        """
        res = self._getRPC(**kwargs).getOperationsFromRMSOpID(rmsOpID)
        if not res["OK"]:
            return res

        operationsJSON = res["Value"]
        try:
            operations, _size = decode(operationsJSON)
            return S_OK(operations)
        except Exception as e:
            return S_ERROR(0, "Exception when decoding the operations json %s" % e)



""" VOMS2CSSyncronizer is a helper class containing the logic for synchronization
    of the VOMS user data with the DIRAC Registry
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from collections import defaultdict

from DIRAC import S_OK, S_ERROR, gLogger, gConfig

from DIRAC.Core.Security.VOMSService import VOMSService
from DIRAC.Core.Utilities.List import fromChar
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import (
    getVOOption,
    getVOMSRoleGroupMapping,
    getUsersInVO,
    getAllUsers,
    getUserOption,
    getGroupsForUser,
)


def _getUserNameFromMail(mail):
    """Utility to construct a reasonable user name from the user mail address

    :param str mail: e-mail address
    :return str: user name
    """

    mailName = mail.split("@")[0].lower()
    if "." in mailName:
        # Most likely the mail contains the full user name
        names = mailName.split(".")
        name = names[0][0] + names[-1].lower()
        return name

    return mailName


def _getUserNameFromDN(dn, vo):
    """Utility to construct a reasonable user name from the user DN
    :param str dn: user DN
    :return str: user name
    """

    shortVO = vo
    if "." in vo:
        vos = vo.split(".")
        if vos[0] == "vo":
            vos = vos[1:]
        if len(vos[-1]) == 2 or vos[-1] == "org":
            vos = vos[:1]
        shortVO = ".".join(vos)

    # Weird case of just a name as DN !
    if "/" not in dn and "CN=" not in dn:
        dn = "CN=" + dn
    entries = dn.split("/")
    entries.reverse()
    for entry in entries:
        if entry:
            # Weird case of no field name !
            if "=" not in entry:
                key, value = "CN", entry
            else:
                key, value = entry.split("=")
            if key.upper() == "CN":
                ind = value.find("(")
                # Strip of possible words in parenthesis in the name
                if ind != -1:
                    value = value[:ind]
                names = value.split()
                if len(names) == 1:
                    nname = names[0].lower()
                    if "." in nname:
                        names = nname.split(".")
                        nname = (names[0][0] + names[-1]).lower()
                    if "-" in nname:
                        names = nname.split("-")
                        nname = (names[0][0] + names[-1]).lower()
                    return nname
                else:
                    robot = False
                    if names[0].lower().startswith("robot"):
                        names.pop(0)
                        robot = True
                    for name in list(names):
                        if name[0].isdigit() or "@" in name:
                            names.pop(names.index(name))
                    if robot:
                        nname = "robot-%s-%s" % (names[-1].lower(), shortVO)
                    else:
                        nname = (names[0][0] + names[-1]).lower()
                        if "." in nname:
                            names = nname.split(".")
                            nname = (names[0][0] + names[-1]).lower()
                    return nname


def _getUserNameFromSurname(name, surname):
    """Construct a reasonable userName from the user name and surname

    :param str name: user name
    :param str surname: user surname
    :return str: constructed user name
    """
    names = name.split()
    initials = ""
    for nn in names:
        initials += nn[0]
    surnames = surname.split()
    result = initials + surnames[-1]
    if len(result) >= 12:
        return result[:11]
    result = result.lower()
    return result


class VOMS2CSSynchronizer(object):
    def __init__(
        self,
        vo,
        autoModifyUsers=True,
        autoAddUsers=True,
        autoDeleteUsers=False,
        autoLiftSuspendedStatus=False,
        syncPluginName=None,
    ):
        """VOMS2CSSynchronizer class constructor

        :param str vo: VO to be synced
        :param boolean autoModifyUsers: flag to automatically modify user data in CS
        :param autoAddUsers: flag to automatically add new users to CS
        :param autoDeleteUsers: flag to automatically delete users from CS if no more in VOMS
        :param autoLiftSuspendedStatus: flag to automatically remove Suspended status in CS
        :param syncPluginName: name of the plugin to validate or extend users' info

        :return: None
        """

        self.log = gLogger.getSubLogger("VOMS2CSSynchronizer")
        self.csapi = CSAPI()
        self.vo = vo
        self.vomsVOName = getVOOption(vo, "VOMSName", "")
        if not self.vomsVOName:
            raise Exception("VOMS name not defined for VO %s" % vo)
        self.adminMsgs = {"Errors": [], "Info": []}
        self.vomsUserDict = {}
        self.autoModifyUsers = autoModifyUsers
        self.autoAddUsers = autoAddUsers
        self.autoDeleteUsers = autoDeleteUsers
        self.autoLiftSuspendedStatus = autoLiftSuspendedStatus
        self.voChanged = False
        self.syncPlugin = None

        if syncPluginName:
            objLoader = ObjectLoader()
            _class = objLoader.loadObject(
                "ConfigurationSystem.Client.SyncPlugins.%sSyncPlugin" % syncPluginName, "%sSyncPlugin" % syncPluginName
            )

            if not _class["OK"]:
                raise Exception(_class["Message"])

            self.syncPlugin = _class["Value"]()

    def syncCSWithVOMS(self):
        """Performs the synchronization of the DIRAC registry with the VOMS data. The resulting
          CSAPI object containing modifications is returned as part of the output dictionary.
          Those changes can be applied by the caller depending on the mode (dry or a real run)


        :return: S_OK with a dictionary containing the results of the synchronization operation
        """
        resultDict = defaultdict(list)

        # Get DIRAC group vs VOMS Role Mappings
        result = getVOMSRoleGroupMapping(self.vo)
        if not result["OK"]:
            return result

        vomsDIRACMapping = result["Value"]["VOMSDIRAC"]
        diracVOMSMapping = result["Value"]["DIRACVOMS"]
        noVOMSGroups = result["Value"]["NoVOMS"]
        noSyncVOMSGroups = result["Value"]["NoSyncVOMS"]

        vomsSrv = VOMSService(self.vo)

        # Get VOMS users
        result = vomsSrv.getUsers()
        if not result["OK"]:
            self.log.error("Could not retrieve user information from VOMS", result["Message"])
            return result

        self.vomsUserDict = result["Value"]
        message = "There are %s user entries in VOMS for VO %s" % (len(self.vomsUserDict), self.vomsVOName)
        self.adminMsgs["Info"].append(message)
        self.log.info("VOMS user entries", message)
        self.log.debug(self.vomsUserDict)

        # Get DIRAC users
        result = self.getVOUserData(self.vo)
        if not result["OK"]:
            return result
        diracUserDict = result["Value"]
        self.adminMsgs["Info"].append(
            "There are %s registered users in DIRAC for VO %s" % (len(diracUserDict), self.vo)
        )
        self.log.info(
            "Users already registered", ": there are %s registered users in DIRAC VO %s" % (len(diracUserDict), self.vo)
        )

        # Find new and obsoleted user DNs
        existingDNs = []
        obsoletedDNs = []
        newDNs = []
        for user in diracUserDict:
            dn = diracUserDict[user]["DN"]
            # We can have users with more than one DN registered
            dnList = fromChar(dn)
            existingDNs.extend(dnList)
            for dn in dnList:
                if dn not in self.vomsUserDict:
                    obsoletedDNs.append(dn)

        for dn in self.vomsUserDict:
            if dn not in existingDNs:
                newDNs.append(dn)

        allDiracUsers = getAllUsers()
        nonVOUserDict = {}
        nonVOUsers = list(set(allDiracUsers) - set(diracUserDict))
        if nonVOUsers:
            result = self.csapi.describeUsers(nonVOUsers)
            if not result["OK"]:
                self.log.error("Could not retrieve CS User description")
                return result
            nonVOUserDict = result["Value"]

        # Process users
        defaultVOGroup = getVOOption(self.vo, "DefaultGroup", "%s_user" % self.vo)
        # If a user is (previously put by hand) in a "QuarantineGroup",
        # then the default group will be ignored.
        # So, this option is only considered for the case of existing users.
        quarantineVOGroup = getVOOption(self.vo, "QuarantineGroup")

        newAddedUserDict = {}
        for dn in self.vomsUserDict:
            newDNForExistingUser = ""
            diracName = ""
            if dn in existingDNs:
                for user in diracUserDict:
                    if dn in fromChar(diracUserDict[user]["DN"]):
                        diracName = user
                        break

            if dn in newDNs:
                # Find if the DN is already registered in the DIRAC CS
                for user in nonVOUserDict:
                    if dn in fromChar(nonVOUserDict[user]["DN"]):
                        diracName = user
                        break

                # Check the nickName in the same VO to see if the user is already registered
                # with another DN
                nickName = self.vomsUserDict[dn].get("nickname")
                if nickName in diracUserDict or nickName in newAddedUserDict:
                    diracName = nickName
                    # This is a flag for adding the new DN to an already existing user
                    newDNForExistingUser = dn

                # We have a real new user
                if not diracName:
                    if nickName:
                        newDiracName = nickName
                    else:
                        newDiracName = self.getUserName(dn)

                    # Do not consider users with Suspended status in VOMS
                    if self.vomsUserDict[dn]["suspended"] or self.vomsUserDict[dn]["certSuspended"]:
                        resultDict["SuspendedUsers"].append(newDiracName)
                        continue

                    # If the chosen user name exists already, append a distinguishing suffix
                    ind = 1
                    trialName = newDiracName
                    while newDiracName in allDiracUsers:
                        # We have a user with the same name but with a different DN
                        newDiracName = "%s_%d" % (trialName, ind)
                        ind += 1

                    # We now have everything to add the new user
                    userDict = {"DN": dn, "CA": self.vomsUserDict[dn]["CA"], "Email": self.vomsUserDict[dn]["mail"]}
                    groupsWithRole = []
                    for role in self.vomsUserDict[dn]["Roles"]:
                        groupList = vomsDIRACMapping.get(role, [])
                        for group in groupList:
                            if group not in noSyncVOMSGroups:
                                groupsWithRole.append(group)
                    userDict["Groups"] = list(set(groupsWithRole + [defaultVOGroup]))

                    # Run the sync plugins for extra info and/or validations
                    if self.syncPlugin:
                        try:
                            self.syncPlugin.verifyAndUpdateUserInfo(newDiracName, userDict)
                        except ValueError as e:
                            self.log.error("Error validating new user", "nickname %s\n error %s" % (newDiracName, e))
                            self.adminMsgs["Errors"].append(
                                "Error validating new user %s: %s\n  %s" % (newDiracName, userDict, e)
                            )
                            continue

                    message = "\n  Added new user %s:\n" % newDiracName
                    for key in userDict:
                        message += "    %s: %s\n" % (key, str(userDict[key]))
                    self.adminMsgs["Info"].append(message)
                    self.voChanged = True
                    if self.autoAddUsers:
                        self.log.info("Adding new user %s: %s" % (newDiracName, str(userDict)))
                        result = self.csapi.modifyUser(newDiracName, userDict, createIfNonExistant=True)
                        if not result["OK"]:
                            self.log.warn("Failed adding new user %s" % newDiracName)
                        resultDict["NewUsers"].append(newDiracName)
                        newAddedUserDict[newDiracName] = userDict
                    continue

            # We have an already existing user
            modified = False
            suspendedInVOMS = self.vomsUserDict[dn]["suspended"] or self.vomsUserDict[dn]["certSuspended"]
            suspendedVOList = getUserOption(diracName, "Suspended", [])
            knownEmail = getUserOption(diracName, "Email", None)
            userDict = {
                "DN": diracUserDict[diracName]["DN"],
                "CA": self.vomsUserDict[dn]["CA"],
                "Email": self.vomsUserDict[dn].get("mail", self.vomsUserDict[dn].get("emailAddress")) or knownEmail,
            }

            # Set Suspended status for the user for this particular VO
            if suspendedInVOMS and self.vo not in suspendedVOList:
                suspendedVOList.append(self.vo)
                userDict["Suspended"] = ",".join(suspendedVOList)
                modified = True

            # Remove the lifted Suspended status
            if not suspendedInVOMS and self.vo in suspendedVOList and self.autoLiftSuspendedStatus:
                newList = []
                for vo in suspendedVOList:
                    if vo != self.vo:
                        newList.append(vo)
                if not newList:
                    newList = ["None"]
                userDict["Suspended"] = ",".join(newList)
                modified = True

            if newDNForExistingUser:
                userDict["DN"] = ",".join([dn, diracUserDict.get(diracName, newAddedUserDict.get(diracName))["DN"]])
                modified = True
            existingGroups = diracUserDict.get(diracName, {}).get("Groups", [])
            nonVOGroups = list(set(existingGroups) - set(diracVOMSMapping))
            groupsWithRole = []
            for role in self.vomsUserDict[dn]["Roles"]:
                groupList = vomsDIRACMapping.get(role, [])
                for group in groupList:
                    if group not in noSyncVOMSGroups:
                        groupsWithRole.append(group)
            keepGroups = nonVOGroups + groupsWithRole
            if not quarantineVOGroup or quarantineVOGroup not in existingGroups:
                keepGroups += [defaultVOGroup]
            if quarantineVOGroup and quarantineVOGroup in existingGroups:
                keepGroups = [quarantineVOGroup]
            for group in existingGroups:
                if group in nonVOGroups:
                    continue
                role = diracVOMSMapping.get(group, "")
                # Among already existing groups for the user keep those without a special VOMS Role
                # because this membership is done by hand in the CS
                if "Role" not in role:
                    keepGroups.append(group)
                # Keep existing groups with no VOMS attribute if any
                if group in noVOMSGroups:
                    keepGroups.append(group)
                # Keep groups for which syncronization with VOMS is forbidden
                if group in noSyncVOMSGroups:
                    keepGroups.append(group)
            userDict["Groups"] = list(set(keepGroups))
            # Merge together groups for the same user but different DNs
            if diracName in newAddedUserDict:
                otherGroups = newAddedUserDict[diracName].get("Groups", [])
                userDict["Groups"] = list(set(keepGroups + otherGroups))
                modified = True
            if not existingGroups and diracName in allDiracUsers:
                groups = getGroupsForUser(diracName)
                if groups["OK"]:
                    self.log.info("Found groups for user %s %s" % (diracName, groups["Value"]))
                    userDict["Groups"] = list(set(groups["Value"] + keepGroups))
                    addedGroups = list(set(userDict["Groups"]) - set(groups["Value"]))
                    modified = True
                    message = "\n  Modified user %s:\n" % diracName
                    message += "    Added to group(s) %s\n" % ",".join(addedGroups)
                    self.adminMsgs["Info"].append(message)

            # Check if something changed before asking CSAPI to modify
            if diracName in diracUserDict:
                message = "\n  Modified user %s:\n" % diracName
                modMsg = ""
                for key in userDict:
                    if key == "Groups":
                        addedGroups = set(userDict[key]) - set(diracUserDict.get(diracName, {}).get(key, []))
                        removedGroups = set(diracUserDict.get(diracName, {}).get(key, [])) - set(userDict[key])
                        if addedGroups:
                            modMsg += "    Added to group(s) %s\n" % ",".join(addedGroups)
                        if removedGroups:
                            modMsg += "    Removed from group(s) %s\n" % ",".join(removedGroups)
                    elif key == "Suspended":
                        if userDict["Suspended"] == "None":
                            modMsg += "    Suspended status removed\n"
                        else:
                            modMsg += "    User Suspended in VOs: %s\n" % userDict["Suspended"]
                    else:
                        oldValue = str(diracUserDict.get(diracName, {}).get(key, ""))
                        if str(userDict[key]) != oldValue:
                            modMsg += "    %s: %s -> %s\n" % (key, oldValue, str(userDict[key]))
                if modMsg:
                    self.adminMsgs["Info"].append(message + modMsg)
                    modified = True

            if self.autoModifyUsers and modified:
                result = self.csapi.modifyUser(diracName, userDict)
                if result["OK"] and result["Value"]:
                    self.log.info("Modified user %s: %s" % (diracName, str(userDict)))
                    self.voChanged = True
                    resultDict["ModifiedUsers"].append(diracName)

        # Check if there are potentially obsoleted users
        oldUsers = set()
        for user in diracUserDict:
            dnSet = set(fromChar(diracUserDict[user]["DN"]))
            if not dnSet.intersection(set(self.vomsUserDict)) and user not in nonVOUserDict:
                existingGroups = diracUserDict.get(user, {}).get("Groups", [])
                nonVOGroups = list(set(existingGroups) - set(diracVOMSMapping))
                removedGroups = list(set(existingGroups) - set(nonVOGroups))
                if removedGroups:
                    self.log.info("Checking user for deletion", "%s: %s" % (user, existingGroups))
                    self.log.info("User has groups in other VOs", "%s: %s" % (user, nonVOGroups))
                    userDict = diracUserDict[user]
                    userDict["Groups"] = nonVOGroups
                    if self.autoModifyUsers:
                        result = self.csapi.modifyUser(user, userDict)
                        if result["OK"] and result["Value"]:
                            self.log.info("Modified user %s: %s" % (user, str(userDict)))
                            self.voChanged = True
                            message = "\n  Modified user %s:\n" % user
                            modMsg = "    Removed from group(s) %s\n" % ",".join(removedGroups)
                            self.adminMsgs["Info"].append(message + modMsg)
                            resultDict["ModifiedUsers"].append(user)
                    continue
                if not any(group in noVOMSGroups for group in existingGroups):
                    oldUsers.add(user)

        # Check for obsoleted DNs
        for user in diracUserDict:
            dnSet = set(fromChar(diracUserDict[user]["DN"]))
            for dn in dnSet:
                if dn in obsoletedDNs and user not in oldUsers:
                    existingGroups = diracUserDict.get(user, {}).get("Groups", [])
                    nonVOGroups = list(set(existingGroups) - set(diracVOMSMapping))
                    if nonVOGroups:
                        self.log.verbose("User has groups in other VOs", "%s: %s" % (user, nonVOGroups))
                        continue
                    self.log.verbose("Modified user %s: dropped DN %s" % (user, dn))
                    if self.autoModifyUsers:
                        userDict = diracUserDict[user]
                        modDNSet = dnSet - set([dn])
                        if modDNSet:
                            userDict["DN"] = ",".join(modDNSet)
                            result = self.csapi.modifyUser(user, userDict)
                            if result["OK"] and result["Value"]:
                                self.log.info("Modified user %s: dropped DN %s" % (user, dn))
                                self.adminMsgs["Info"].append("Modified user %s: dropped DN %s" % (user, dn))
                                self.voChanged = True
                                resultDict["ModifiedUsers"].append(diracName)
                        else:
                            oldUsers.add(user)

        if oldUsers:
            self.voChanged = True
            if self.autoDeleteUsers:
                self.log.info("The following users will be deleted: %s" % str(oldUsers))
                result = self.csapi.deleteUsers(oldUsers)
                if result["OK"]:
                    self.adminMsgs["Info"].append("The following users are deleted from CS:\n  %s\n" % str(oldUsers))
                    resultDict["DeletedUsers"] = oldUsers
                else:
                    self.adminMsgs["Errors"].append("Error in deleting users from CS:\n  %s" % str(oldUsers))
                    self.log.error("Error while user deletion from CS", result)
            else:
                self.adminMsgs["Info"].append(
                    "The following users to be checked for deletion:\n\t%s" % "\n\t".join(sorted(oldUsers))
                )
                self.log.info("The following users to be checked for deletion:", "\n\t".join(sorted(oldUsers)))

        resultDict["CSAPI"] = self.csapi
        resultDict["AdminMessages"] = self.adminMsgs
        resultDict["VOChanged"] = self.voChanged
        return S_OK(resultDict)

    def getVOUserData(self, refreshFlag=False):
        """Get a report for users of a given VO

        :param bool refreshFlag: flag to indicate that the configuration must be refreshed
                                 before looking up user data
        :return: S_OK/S_ERROR, Value = user description dictionary
        """
        if refreshFlag:
            gConfig.forceRefresh()

        # Get DIRAC users
        diracUsers = getUsersInVO(self.vo)
        if not diracUsers:
            return S_ERROR("No VO users found for %s" % self.vo)

        if refreshFlag:
            result = self.csapi.downloadCSData()
            if not result["OK"]:
                return result
        result = self.csapi.describeUsers(diracUsers)
        if not result["OK"]:
            self.log.error("Could not retrieve CS User description")
        return result

    def getVOUserReport(self):
        """Get a report string with the current status of the DIRAC Registry for the
            Virtual Organization

        :return: S_OK with the report string as Value
        """

        result = self.getVOUserData(refreshFlag=True)
        if not result["OK"]:
            return result

        userDict = result["Value"]

        # Get DIRAC group vs VOMS Role Mappings
        result = getVOMSRoleGroupMapping(self.vo)
        if not result["OK"]:
            return result

        diracVOMSMapping = result["Value"]["DIRACVOMS"]
        records = []
        groupDict = defaultdict(int)
        multiDNUsers = {}
        suspendedUsers = []
        for user in userDict:
            for group in userDict[user]["Groups"]:
                groupDict[group] += 1
            dnList = fromChar(userDict[user]["DN"])
            if len(dnList) > 1:
                multiDNUsers[user] = dnList
            if userDict[user].get("Status", "Active") == "Suspended":
                suspendedUsers.append(user)

        for group in diracVOMSMapping:
            records.append((group, str(groupDict[group]), diracVOMSMapping.get(group, "")))

        fields = ["Group", "Number of users", "VOMS Role"]
        output = printTable(fields, records, sortField="Group", printOut=False, numbering=False)

        if multiDNUsers:
            output += "\nUsers with multiple DNs:\n"
            for user in multiDNUsers:
                output += "  %s:\n" % user
                for dn in multiDNUsers[user]:
                    output += "    %s\n" % dn

        if suspendedUsers:
            output += "\n%d suspended users:\n" % len(suspendedUsers)
            output += "  %s" % ",".join(suspendedUsers)

        return S_OK(output)

    def getUserName(self, dn):
        """Utility to construct user name

        :param str dn: user DN
        :return str: user name
        """
        name = self.vomsUserDict[dn].get("name")
        surname = self.vomsUserDict[dn].get("surname")
        if name and surname:
            surnameName = _getUserNameFromSurname(name, surname)
            return surnameName

        dnName = _getUserNameFromDN(dn, self.vo)

        # If robot, take the dn based name
        if dnName.startswith("robot"):
            return dnName

        # Is mailName reasonable ?
        mail = self.vomsUserDict[dn]["mail"]
        if mail:
            mailName = _getUserNameFromMail(mail)
            if len(mailName) > 5 and mailName.isalpha():
                return mailName

        # dnName too long
        if len(dnName) >= 12:
            dnName = dnName[:11]

        # May be the mail name is still more reasonable
        if mail and len(dnName) < len(mailName) and mailName.isalpha():
            return mailName

        return dnName



#!/usr/bin/env python
"""
Add files to an existing transformation
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("TransID: transformation ID")
    Script.registerArgument(("LFN: LFN", "FileName: file containing LFNs"))
    Script.parseCommandLine()

    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    tID, inputFileName = Script.getPositionalArgs(group=True)

    # get arguments
    lfns = []
    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        inputFile.close()
        lfns.extend([lfn.strip() for lfn in string.splitlines()])
    else:
        lfns.append(inputFileName)

    tc = TransformationClient()
    res = tc.addFilesToTransformation(tID, lfns)  # Files added here

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(2)

    successfullyAdded = 0
    alreadyPresent = 0
    for lfn, message in res["Value"]["Successful"].items():
        if message == "Added":
            successfullyAdded += 1
        elif message == "Present":
            alreadyPresent += 1

    if successfullyAdded > 0:
        DIRAC.gLogger.notice("Successfully added %d files" % successfullyAdded)
    if alreadyPresent > 0:
        DIRAC.gLogger.notice("Already present %d files" % alreadyPresent)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()
